# Copyright(C) 2019-2021 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import os
from datetime import datetime
from woob.tools.log import getLogger

__all__ = ['LOGGER', 'replace', 'time_buffer']


LOGGER = getLogger('woob.config')


try:
    from os import replace
except ImportError:
    def replace(src, dst, *args, **kwargs):
        if os.name != 'posix':
            try:
                os.remove(dst)
            except OSError:
                pass
        return os.rename(src, dst, *args, **kwargs)


def time_buffer(since_seconds=None, last_run=True, logger=False):
    def decorator_time_buffer(func):
        def wrapper_time_buffer(*args, **kwargs):
            since_seconds = kwargs.pop('since_seconds', None)
            if since_seconds is None:
                since_seconds = decorator_time_buffer.since_seconds
            if logger:
                logger.debug('Time buffer for %r of %s. Last run %s.'
                             % (func, since_seconds, decorator_time_buffer.last_run))
            if since_seconds and decorator_time_buffer.last_run:
                if (datetime.now() - decorator_time_buffer.last_run).seconds < since_seconds:
                    if logger:
                        logger.debug('Too soon to run %r, ignore.' % func)
                    return
            if logger:
                logger.debug('Run %r and record' % func)
            res = func(*args, **kwargs)
            decorator_time_buffer.last_run = datetime.now()
            return res

        decorator_time_buffer.since_seconds = since_seconds
        decorator_time_buffer.last_run = datetime.now() if last_run is True else last_run

        return wrapper_time_buffer
    return decorator_time_buffer



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from copy import deepcopy

from .config.yamlconfig import YamlConfig


class IStorage:
    def load(self, what, name, default={}):
        """
        Load data from storage.
        """
        raise NotImplementedError()

    def save(self, what, name):
        """
        Write changes in storage on the disk.
        """
        raise NotImplementedError()

    def set(self, what, name, *args):
        """
        Set data in a path.
        """
        raise NotImplementedError()

    def delete(self, what, name, *args):
        """
        Delete a value or a path.
        """
        raise NotImplementedError()

    def get(self, what, name, *args, **kwargs):
        """
        Get a value or a path.
        """
        raise NotImplementedError()


class StandardStorage(IStorage):
    def __init__(self, path):
        self.config = YamlConfig(path)
        self.config.load()

    def load(self, what, name, default={}):
        d = {}
        if what not in self.config.values:
            self.config.values[what] = {}
        else:
            d = self.config.values[what].get(name, {})

        self.config.values[what][name] = deepcopy(default)
        self.config.values[what][name].update(d)

    def save(self, what, name):
        self.config.save()

    def set(self, what, name, *args):
        self.config.set(what, name, *args)

    def delete(self, what, name, *args):
        self.config.delete(what, name, *args)

    def get(self, what, name, *args, **kwargs):
        return self.config.get(what, name, *args, **kwargs)



# Copyright(C) 2010-2012 Romain Bignon, Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import logging
import optparse
from optparse import OptionGroup, OptionParser
from datetime import datetime
import os
import sys
import warnings

from woob.capabilities.base import ConversionWarning, BaseObject
from woob.core import Woob, CallErrors
from woob.core.backendscfg import BackendsConfig
from woob.tools.config.iconfig import ConfigError
from woob.exceptions import FormFieldConversionWarning
from woob.tools.log import createColoredFormatter, getLogger, DEBUG_FILTERS, settings as log_settings
from woob.tools.misc import guess_encoding

from .results import ResultsConditionError

__all__ = ['Application']


class MoreResultsAvailable(Exception):
    pass


class ApplicationStorage:
    def __init__(self, name, storage):
        self.name = name
        self.storage = storage

    def set(self, *args):
        if self.storage:
            return self.storage.set('applications', self.name, *args)

    def delete(self, *args):
        if self.storage:
            return self.storage.delete('applications', self.name, *args)

    def get(self, *args, **kwargs):
        if self.storage:
            return self.storage.get('applications', self.name, *args, **kwargs)
        else:
            return kwargs.get('default', None)

    def load(self, default):
        if self.storage:
            return self.storage.load('applications', self.name, default)

    def save(self):
        if self.storage:
            return self.storage.save('applications', self.name)


class Application:
    """
    Base application.

    This class can be herited to have some common code within woob
    applications.
    """

    # ------ Class attributes --------------------------------------

    APPNAME = ''
    """Application name"""

    OLD_APPNAME = ""

    CONFDIR = None
    """Configuration and work directory (if None, use the Woob instance one)"""

    CONFIG = {}
    """Default configuration dict (can only contain key/values)"""

    STORAGE = {}
    """Default storage tree"""

    SYNOPSIS = 'Usage: %prog [-h] [-dqv] [-b backends] ...\n'
    SYNOPSIS += '       %prog [--help] [--version]'
    """Synopsis"""

    DESCRIPTION = None
    """Description"""

    VERSION = None
    """Version"""

    COPYRIGHT = None
    """Copyright"""

    DEBUG_FILTER = 2
    """Verbosity of DEBUG"""

    stdin = sys.stdin
    stdout = sys.stdout
    stderr = sys.stderr

    # ------ Abstract methods --------------------------------------
    def create_woob(self):
        return Woob()

    def _get_completions(self):
        """
        Overload this method in subclasses if you want to enrich shell completion.
        @return  a set object
        """
        return set()

    def _handle_options(self):
        """
        Overload this method in application type subclass
        if you want to handle options defined in subclass constructor.
        """
        pass

    def add_application_options(self, group):
        """
        Overload this method if your application needs extra options.

        These options will be displayed in an option group.
        """
        pass

    def handle_application_options(self):
        """
        Overload this method in your application if you want to handle options defined in add_application_options.
        """
        pass

    # ------ Application methods -------------------------------

    def __init__(self, option_parser=None):
        super(Application, self).__init__()
        self.encoding = self.guess_encoding()
        self.logger = getLogger("woob.applications.%s" % self.APPNAME)
        self.woob = self.create_woob()
        if self.CONFDIR is None:
            self.CONFDIR = self.woob.workdir
        self.config = None
        self.options = None
        self.condition = None
        self.storage = None
        if option_parser is None:
            self._parser = OptionParser(self.SYNOPSIS, version=self._get_optparse_version())
        else:
            self._parser = option_parser
        if self.DESCRIPTION:
            self._parser.description = self.DESCRIPTION
        app_options = OptionGroup(self._parser, '%s Options' % self.APPNAME.capitalize())
        self.add_application_options(app_options)
        if len(app_options.option_list) > 0:
            self._parser.add_option_group(app_options)
        self._parser.add_option('-b', '--backends', help='what backend(s) to enable (comma separated)')
        self._parser.add_option('-e', '--exclude-backends', help='what backend(s) to exclude (comma separated)')

        self._parser.add_option('-I', '--insecure', action='store_true', help='do not validate SSL')
        self._parser.add_option('--nss', action='store_true', help='Use NSS instead of OpenSSL')
        self._parser.add_option('--force-ipv4', action='store_const', help='Force IPv4', const=4, dest="ipversion")
        self._parser.add_option('--force-ipv6', action='store_const', help='Force IPv6', const=6, dest="ipversion")

        logging_options = OptionGroup(self._parser, 'Logging Options')
        logging_options.add_option('-d', '--debug', action='count', help='display debug messages. Set up it twice to more verbosity', default=0)
        logging_options.add_option('-q', '--quiet', action='store_true', help='display only error messages')
        logging_options.add_option('-v', '--verbose', action='store_true', help='display info messages')
        logging_options.add_option('--logging-file', action='store', type='string', dest='logging_file', help='file to save logs')
        logging_options.add_option('-a', '--save-responses', action='store_true', help='save every response')
        logging_options.add_option('--export-session', action='store_true', help='log browser session cookies after login')
        self._parser.add_option_group(logging_options)
        self._parser.add_option('--shell-completion', action='store_true', help=optparse.SUPPRESS_HELP)
        self._is_default_count = True

    def guess_encoding(self, stdio=None):
        return guess_encoding(stdio or self.stdout)

    def deinit(self):
        self.woob.want_stop()
        self.woob.deinit()

    def _get_preferred_path(self, preferred, legacy):
        try:
            os.lstat(preferred)
        except FileNotFoundError:
            pass
        else:
            return preferred

        if os.path.exists(legacy):
            self.logger.debug("legacy %r can be renamed", legacy)

            try:
                os.rename(legacy, preferred)
            except IOError as exc:
                self.logger.warning("can't rename legacy %r: %s", legacy, exc)
                return legacy

        return preferred

    def create_storage(self, path=None, klass=None, localonly=False):
        """
        Create a storage object.

        :param path: An optional specific path
        :type path: :class:`str`
        :param klass: What class to instance
        :type klass: :class:`woob.tools.storage.IStorage`
        :param localonly: If True, do not set it on the :class:`Woob` object.
        :type localonly: :class:`bool`
        :rtype: :class:`woob.tools.storage.IStorage`
        """
        if klass is None:
            from woob.tools.storage import StandardStorage
            klass = StandardStorage

        if path is None:
            path = os.path.join(self.CONFDIR, self.APPNAME + '.storage')
            if self.OLD_APPNAME:
                # compatibility for old, non-woob names
                path = self._get_preferred_path(
                    path, os.path.join(self.CONFDIR, self.OLD_APPNAME + '.storage')
                )
        elif os.path.sep not in path:
            path = os.path.join(self.CONFDIR, path)

        storage = klass(path)
        self.storage = ApplicationStorage(self.APPNAME, storage)
        self.storage.load(self.STORAGE)

        if not localonly:
            self.woob.storage = storage

        return storage

    def load_config(self, path=None, klass=None):
        """
        Load a configuration file and get his object.

        :param path: An optional specific path
        :type path: :class:`str`
        :param klass: What class to instance
        :type klass: :class:`woob.tools.config.iconfig.IConfig`
        :rtype: :class:`woob.tools.config.iconfig.IConfig`
        """
        if klass is None:
            from woob.tools.config.iniconfig import INIConfig
            klass = INIConfig

        if path is None:
            path = os.path.join(self.CONFDIR, self.APPNAME)
            if self.OLD_APPNAME:
                # compatibility for old, non-woob names
                path = self._get_preferred_path(
                    path, os.path.join(self.CONFDIR, self.OLD_APPNAME)
                )
        elif os.path.sep not in path:
            path = os.path.join(self.CONFDIR, path)

        self.config = klass(path)
        self.config.load(self.CONFIG)

        if int(self.config.get('use_nss', default=0)):
            # TODO Deprecated
            print('Key "use_nss" is deprecated and will be ignored', file=sys.stderr)

        if int(self.config.get('export_session', default=0)):
            log_settings['export_session'] = True

    def main(self, argv):
        """
        Main method

        Called by run
        """
        raise NotImplementedError()

    def load_backends(self, caps=None, names=None, exclude=None, *args, **kwargs):
        if names is None and self.options.backends:
            names = self.options.backends.split(',')
        if exclude is None and self.options.exclude_backends:
            exclude = self.options.exclude_backends.split(',')
        loaded = self.woob.load_backends(caps, names, exclude=exclude, *args, **kwargs)
        if not loaded:
            logging.info('No backend loaded')
        return loaded

    def _get_optparse_version(self):
        version = None
        if self.VERSION:
            if self.COPYRIGHT:
                copyright = self.COPYRIGHT.replace('YEAR', '%d' % datetime.today().year)
                version = 'Woob %s v%s %s' % (self.APPNAME, self.VERSION, copyright)
            else:
                version = 'Woob %s v%s' % (self.APPNAME, self.VERSION)
        return version

    def _do_complete_obj(self, backend, fields, obj):
        if not obj:
            return obj
        if not isinstance(obj, BaseObject):
            return obj

        obj.backend = backend.name
        if fields is None or len(fields) > 0:
            obj = backend.fillobj(obj, fields) or obj
        return obj

    def _do_complete_iter(self, backend, count, fields, res):
        modif = 0

        for i, sub in enumerate(res):
            sub = self._do_complete_obj(backend, fields, sub)
            if self.condition and self.condition.limit and \
               self.condition.limit == i:
                return

            if self.condition and not self.condition.is_valid(sub):
                modif += 1
            else:
                if count and i - modif == count:
                    if self._is_default_count:
                        raise MoreResultsAvailable()
                    else:
                        return
                yield sub

    def _do_complete(self, backend, count, selected_fields, function, *args, **kwargs):
        assert count is None or count > 0
        if callable(function):
            res = function(backend, *args, **kwargs)
        else:
            res = getattr(backend, function)(*args, **kwargs)

        if hasattr(res, '__iter__') and not isinstance(res, (bytes, str)):
            return self._do_complete_iter(backend, count, selected_fields, res)
        else:
            return self._do_complete_obj(backend, selected_fields, res)

    def bcall_error_handler(self, backend, error, backtrace):
        """
        Handler for an exception inside the CallErrors exception.

        This method can be overridden to support more exceptions types.
        """

        # Ignore this error.
        if isinstance(error, MoreResultsAvailable):
            return False

        print('Error(%s): %s' % (backend.name, error), file=self.stderr)
        if logging.root.level <= logging.DEBUG:
            print(backtrace, file=self.stderr)
        else:
            return True

    def bcall_errors_handler(self, errors, debugmsg='Use --debug option to print backtraces', ignore=()):
        """
        Handler for the CallErrors exception.

        It calls `bcall_error_handler` for each error.

        :param errors: Object containing errors from backends
        :type errors: :class:`woob.core.bcall.CallErrors`
        :param debugmsg: Default message asking to enable the debug mode
        :type debugmsg: :class:`str`
        :param ignore: Exceptions to ignore
        :type ignore: tuple[:class:`Exception`]
        """
        err = 0

        ask_debug_mode = False
        for backend, error, backtrace in errors.errors:
            if isinstance(error, ignore):
                continue
            elif self.bcall_error_handler(backend, error, backtrace):
                ask_debug_mode = True

            if not isinstance(error, MoreResultsAvailable):
                err = 1

        if ask_debug_mode:
            print(debugmsg, file=self.stderr)

        return err

    def _shell_completion_items(self):
        items = set()
        for ol in [self._parser.option_list] + [og.option_list for og in self._parser.option_groups]:
            for option in ol:
                if option.help is not optparse.SUPPRESS_HELP:
                    items.update(str(option).split('/'))
        items.update(self._get_completions())
        return items

    def parse_args(self, args):
        self.options, args = self._parser.parse_args(args)

        if self.options.shell_completion:
            items = self._shell_completion_items()
            print(' '.join(items))
            sys.exit(0)

        if self.options.debug >= self.DEBUG_FILTER:
            level = DEBUG_FILTERS
        elif self.options.debug or self.options.save_responses:
            level = logging.DEBUG
        elif self.options.verbose:
            level = logging.INFO
        elif self.options.quiet:
            level = logging.ERROR
        else:
            level = logging.WARNING
        if self.options.insecure:
            log_settings['ssl_insecure'] = True
        if self.options.nss:
            print('--nss is deprecated and will be removed', file=sys.stderr)
        if self.options.ipversion:
            self.setup_ipversion()

        # this only matters to developers
        if not self.options.debug and not self.options.save_responses:
            warnings.simplefilter('ignore', category=ConversionWarning)
            warnings.simplefilter('ignore', category=FormFieldConversionWarning)
        else:
            warnings.simplefilter('default')

        handlers = []

        if self.options.save_responses:
            import tempfile
            responses_dirname = tempfile.mkdtemp(prefix='woob_session_')
            print('Debug data will be saved in this directory: %s' % responses_dirname, file=self.stderr)
            log_settings['responses_dirname'] = responses_dirname
            handlers.append(self.create_logging_file_handler(os.path.join(responses_dirname, 'debug.log')))

        if self.options.export_session:
            log_settings['export_session'] = True

        # file logger
        if self.options.logging_file:
            handlers.append(self.create_logging_file_handler(self.options.logging_file))
        else:
            handlers.append(self.create_default_logger())

        self.setup_logging(level, handlers)

        self._handle_options()
        self.handle_application_options()

        return args

    @classmethod
    def create_default_logger(cls):
        # stderr logger
        format = '%(asctime)s:%(levelname)s:%(name)s:' + cls.VERSION +\
                 ':%(filename)s:%(lineno)d:%(funcName)s %(message)s'
        handler = logging.StreamHandler(cls.stderr)
        if os.getenv("NO_COLOR") is None:
            handler.setFormatter(createColoredFormatter(cls.stderr, format))
        return handler

    @classmethod
    def setup_logging(cls, level, handlers):
        logging.root.handlers = []

        logging.root.setLevel(level)
        for handler in handlers:
            logging.root.addHandler(handler)

    def setup_ipversion(self):
        import socket
        import requests.packages.urllib3.util.connection
        import urllib3.util.connection

        for module in (requests.packages.urllib3.util.connection, urllib3.util.connection):
            family = socket.AF_INET
            if self.options.ipversion == 6 and module.HAS_IPV6:
                family = socket.AF_INET6  # force ipv6 only if it is available

            # unfortunately no other way than monkey-patching
            # XXX though probably overkill, we need that to each `module`
            module.allowed_gai_family = lambda family=family: family

    def create_logging_file_handler(self, filename):
        try:
            stream = open(os.path.expanduser(filename), 'w')
        except IOError as e:
            self.logger.error('Unable to create the logging file: %s' % e)
            sys.exit(1)
        else:
            format = '%(asctime)s:%(levelname)s:%(name)s:' + self.VERSION +\
                     ':%(filename)s:%(lineno)d:%(funcName)s %(message)s'
            handler = logging.StreamHandler(stream)
            handler.setFormatter(logging.Formatter(format))
            return handler

    @classmethod
    def run(cls, args=None):
        """
        This static method can be called to run the application.

        It creates the application object, handles options, setups logging, calls
        the main() method, and catches common exceptions.

        You can't do anything after this call, as it *always* finishes with
        a call to sys.exit().

        For example:

        >>> from woob.application.myapplication import MyApplication
        >>> MyApplication.run()
        """

        cls.setup_logging(logging.INFO, [cls.create_default_logger()])

        if args is None:
            args = list(sys.argv)

        try:
            app = cls()
        except BackendsConfig.WrongPermissions as e:
            print(e, file=cls.stderr)
            sys.exit(1)

        try:
            try:
                args = app.parse_args(args)
                sys.exit(app.main(args))
            except KeyboardInterrupt:
                print('Program killed by SIGINT', file=cls.stderr)
                sys.exit(0)
            except EOFError:
                sys.exit(0)
            except ConfigError as e:
                print('Configuration error: %s' % e, file=cls.stderr)
                sys.exit(1)
            except CallErrors as e:
                try:
                    ret = app.bcall_errors_handler(e)
                except KeyboardInterrupt:
                    pass
                else:
                    sys.exit(ret)
                sys.exit(1)
            except ResultsConditionError as e:
                print('%s' % e, file=cls.stderr)
                sys.exit(1)
        finally:
            app.deinit()



# Copyright(C) 2013 Julien Veyssier
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import Capability, BaseObject, StringField, IntField, Field
from .date import DateField


__all__ = ['Movie', 'Person', 'CapCinema']


class Movie(BaseObject):
    """
    Movie object.
    """
    original_title    = StringField('Original title of the movie')
    other_titles      = Field('Titles in other countries', list)
    release_date      = DateField('Release date of the movie')
    all_release_dates = StringField('Release dates list of the movie')
    duration          = IntField('Duration of the movie in minutes')
    short_description = StringField('Short description of the movie')
    genres            = Field('Genres of the movie', list)
    pitch             = StringField('Short story description of the movie')
    country           = StringField('Origin country of the movie')
    note              = StringField('Notation of the movie')
    roles             = Field('Lists of Persons related to the movie indexed by roles', dict)
    thumbnail_url     = StringField('Url of movie thumbnail')

    def __init__(self, id, original_title, url=None):
        super(Movie, self).__init__(id, url)
        self.original_title = original_title

    def get_roles_by_person_name(self, name):
        for role in self.roles.keys():
            if name.lower() in [person[1].lower() for person in self.roles[role]]:
                return role
        return None

    def get_roles_by_person_id(self, id):
        result = []
        for role in self.roles.keys():
            if id in [person[0] for person in self.roles[role]]:
                result.append(role)

        return result


class Person(BaseObject):
    """
    Person object.
    """
    name              = StringField('Star name of a person')
    real_name         = StringField('Real name of a person')
    birth_date        = DateField('Birth date of a person')
    death_date        = DateField('Death date of a person')
    birth_place       = StringField('City and country of birth of a person')
    gender            = StringField('Gender of a person')
    nationality       = StringField('Nationality of a person')
    short_biography   = StringField('Short biography of a person')
    biography         = StringField('Full biography of a person')
    short_description = StringField('Short description of a person')
    roles             = Field('Lists of movies related to the person indexed by roles', dict)
    thumbnail_url     = StringField('Url of person thumbnail')

    def __init__(self, id, name, url=None):
        super(Person, self).__init__(id, url)
        self.name = name

    def get_roles_by_movie_title(self, title):
        for role in self.roles.keys():
            for mt in [movie[1] for movie in self.roles[role]]:
                # title we have is included ?
                if title.lower() in mt.lower():
                    return role
        return None

    def get_roles_by_movie_id(self, id):
        result = []
        for role in self.roles.keys():
            if id in [movie[0] for movie in self.roles[role]]:
                result.append(role)

        return result


class CapCinema(Capability):
    """
    Cinema databases.
    """

    def iter_movies(self, pattern):
        """
        Search movies and iterate on results.

        :param pattern: pattern to search
        :type pattern: str
        :rtype: iter[:class:`Movies`]
        """
        raise NotImplementedError()

    def get_movie(self, _id):
        """
        Get a movie object from an ID.

        :param _id: ID of movie
        :type _id: str
        :rtype: :class:`Movie`
        """
        raise NotImplementedError()

    def get_movie_releases(self, _id, country=None):
        """
        Get a list of a movie releases from an ID.

        :param _id: ID of movie
        :type _id: str
        :rtype: :class:`String`
        """
        raise NotImplementedError()

    def iter_movie_persons(self, _id, role=None):
        """
        Get the list of persons who are related to a movie.

        :param _id: ID of movie
        :type _id: str
        :rtype: iter[:class:`Person`]
        """
        raise NotImplementedError()

    def iter_persons(self, pattern):
        """
        Search persons and iterate on results.

        :param pattern: pattern to search
        :type pattern: str
        :rtype: iter[:class:`persons`]
        """
        raise NotImplementedError()

    def get_person(self, _id):
        """
        Get a person object from an ID.

        :param _id: ID of person
        :type _id: str
        :rtype: :class:`Person`
        """
        raise NotImplementedError()

    def iter_person_movies(self, _id, role=None):
        """
        Get the list of movies related to a person.

        :param _id: ID of person
        :type _id: str
        :rtype: iter[:class:`Movie`]
        """
        raise NotImplementedError()

    def iter_person_movies_ids(self, _id):
        """
        Get the list of movie ids related to a person.

        :param _id: ID of person
        :type _id: str
        :rtype: iter[str]
        """
        raise NotImplementedError()

    def iter_movie_persons_ids(self, _id):
        """
        Get the list of person ids related to a movie.

        :param _id: ID of movie
        :type _id: str
        :rtype: iter[str]
        """
        raise NotImplementedError()

    def get_person_biography(self, id):
        """
        Get the person full biography.

        :param _id: ID of person
        :type _id: str
        :rtype: str
        """
        raise NotImplementedError()



# Copyright(C) 2022      Budget Insight
#
# This file is part of weboob.
#
# weboob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# weboob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with weboob. If not, see <http://www.gnu.org/licenses/>.

import re

from woob.browser.filters.base import Filter, FilterError, debug

__all__ = [
    'SortCodeAccountNumber', 'SortCodeAccountNumberFormatError',
    'canonicalize_sort_code_account_number',
]

STRICT_SORTCODE_ACCOUNT_NUMBER_RE = re.compile(r'\d{14}')
SORTCODE_ACCOUNT_NUMBER_RE = re.compile(
    r'\s*(\d{2})-?(\d{2})-?(\d{2})\s*(\d{4})?\s*(\d{4})\s*',
)


def canonicalize_sort_code_account_number(arg):
    """Return the canonical form of the given sort code account number.

    This function also checks if the provided sort code account number
    is valid, and raises a ValueError in case it is not.
    """
    m = SORTCODE_ACCOUNT_NUMBER_RE.fullmatch(arg)
    if not m:
        raise ValueError(f'{arg!r} is not a valid Sortcode account number')

    return ''.join(m.group(i) for i in range(1, 6))


def is_sort_code_account_number_valid(arg):
    """Return whether the provided sort code account number is valid or not.

    Since this function might be used to check if we want to store the
    provided value or not, we don't want non-canonical values to be validated
    here, and only validate canonical sort code account number representations.
    """
    return bool(STRICT_SORTCODE_ACCOUNT_NUMBER_RE.fullmatch(arg))


class SortCodeAccountNumberFormatError(FilterError):
    pass


class SortCodeAccountNumber(Filter):
    """Return the input only if it is a valid Sortcode account number."""
    @debug()
    def filter(self, code):
        try:
            return canonicalize_sort_code_account_number(code)
        except ValueError:
            return self.default_or_raise(SortCodeAccountNumberFormatError(
                '%r is not a valid sortcode account number, and no ' % code
                + 'default value was set.',
            ))



# Copyright(C) 2009-2012  Jeremy Monnet
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.capabilities.library import CapBook, Book
from woob.tools.application.repl import ReplApplication
from woob.tools.application.formatters.iformatter import PrettyFormatter

__all__ = ['AppBooks']


class RentedListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'date', 'author', 'name', 'late')

    RED = '[1;31m'

    def get_title(self, obj):
        s = '%s — %s (%s)' % (obj.author, obj.name, obj.date)
        if obj.late:
            s += ' %sLATE!%s' % (self.RED, self.NC)
        return s


class AppBooks(ReplApplication):
    APPNAME = 'books'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2012-YEAR Jeremy Monnet'
    CAPS = CapBook
    DESCRIPTION = "Console application allowing to list your books rented or booked at the library, " \
                  "book and search new ones, get your booking history (if available)."
    SHORT_DESCRIPTION = "manage rented books"
    EXTRA_FORMATTERS = {'rented_list':   RentedListFormatter,
                        }
    DEFAULT_FORMATTER = 'table'
    COMMANDS_FORMATTERS = {'ls':          'rented_list',
                           'list':        'rented_list',
                           'rented':      'rented_list',
                          }

    COLLECTION_OBJECTS = (Book, )

    def do_renew(self, id):
        """
        renew ID

        Renew a book
        """

        id, backend_name = self.parse_id(id)
        if not id:
            print('Error: please give a book ID (hint: use ls command)', file=self.stderr)
            return 2
        names = (backend_name,) if backend_name is not None else None

        for renew in self.do('renew_book', id, backends=names):
            self.format(renew)

    def do_rented(self, args):
        """
        rented

        List rented books
        """

        for book in self.do('iter_rented', backends=None):
            self.format(book)

    def do_search(self, pattern):
        """
        search PATTERN

        Search books
        """
        for book in self.do('search_books', pattern):
            self.format(book)



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import List

from .base import (
    Capability, BaseObject, StringField, IntField, Field, UserError,
    NotLoadedType, NotAvailableType, NotLoaded
)


__all__ = ['AccountRegisterError', 'Account', 'StatusField', 'CapAccount', 'CapCredentialsCheck']


class AccountRegisterError(UserError):
    """
    Raised when there is an error during registration.
    """


class Account(BaseObject):
    """
    Describe an account and its properties.
    """
    login =         StringField('Login')
    password =      StringField('Password')
    properties =    Field('List of key/value properties', dict)

    def __init__(
        self,
        id: str = '',
        url: str | NotLoadedType | NotAvailableType = NotLoaded
    ):
        super().__init__(id, url)


class StatusField(BaseObject):
    """
    Field of an account staeobjectus.
    """
    FIELD_TEXT    = 0x001     # the value is a long text
    FIELD_HTML    = 0x002     # the value is HTML formated

    key =           StringField('Key')
    label =         StringField('Label')
    value =         StringField('Value')
    flags =         IntField('Flags')

    def __init__(
        self,
        key: str,
        label: str,
        value: str,
        flags: int = 0,
        url: str | NotLoadedType | NotAvailableType = NotLoaded
    ):
        super().__init__(key, url)
        self.__setattr__('key', key)
        self.__setattr__('label', label)
        self.__setattr__('value', value)
        self.__setattr__('flags', flags)


class CapAccount(Capability):
    """
    Capability for websites when you can create and manage accounts.

    :var ACCOUNT_REGISTER_PROPERTIES: This class constant may be a list of
                                      :class:`woob.tools.value.Value` objects.
                                      If the value remains None, woob considers
                                      that :func:`register_account` isn't supported.
    """
    ACCOUNT_REGISTER_PROPERTIES = None

    @staticmethod
    def register_account(account: Account):
        """
        Register an account on website

        This is a static method, it would be called even if the backend is
        instancied.

        :param account: describe the account to create
        :type account: :class:`Account`
        :raises: :class:`AccountRegisterError`
        """
        raise NotImplementedError()

    def confirm_account(self, mail: str):
        """
        From an email go to the confirm link.
        """
        raise NotImplementedError()

    def get_account(self) -> Account:
        """
        Get the current account.
        """
        raise NotImplementedError()

    def update_account(self, account: Account):
        """
        Update the current account.
        """
        raise NotImplementedError()

    def get_account_status(self) -> List[StatusField]:
        """
        Get status of the current account.

        :returns: a list of fields
        :rtype: list[StatusField]
        """
        raise NotImplementedError()


class CapCredentialsCheck(Capability):
    def check_credentials(self) -> bool:
        """
        Its purpose is to check that the credentials (in the config) are valid
        for the module's website, API, whatever requires credentials to be
        accessed.
        """
        raise NotImplementedError()



# Copyright(C) 2022 Budget Insight
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

# flake8: compatible

from __future__ import annotations

from typing import Dict
from urllib.parse import parse_qsl, urlencode, urlparse

from woob.tools.misc import NO_DEFAULT, NoDefaultType

__all__ = [
    'get_url_fragment_param', 'get_url_fragment_params',
    'get_url_param', 'get_url_params', 'get_url_with_params',
]


def get_url_param(
    url: str, name: str, *,
    default: str | NoDefaultType | None = NO_DEFAULT,
) -> str | None:
    """Get a specific query parameter from an URL.

    :param url: The URL to get the parameter from.
    :param name: The name of the query parameter to get.
    :param default: The default value, as a string or None.
                    Should not be set if the function should raise an exception
                    in cases where no value could be obtained using this URL
                    and name.
    """
    parsed_url = urlparse(url)
    params = dict(parse_qsl(parsed_url.query, keep_blank_values=True))
    if name not in params:
        if default is NO_DEFAULT:
            raise ValueError(
                f'URL {url!r} has no query parameter named {name!r}.',
            )

        return default

    return params[name]


def get_url_fragment_param(
    url: str, name: str, *,
    default: str | NoDefaultType | None = NO_DEFAULT,
) -> str | None:
    """Get a specific fragment parameter from an URL.

    Note that this function is only for cases where the fragment is encoded
    the same way as a query string, e.g. 'https://example.org/#a=b&c=d'.

    :param url: The URL to get the fragment parameter from.
    :param name: The name of the fragment parameter to get.
    :param default: The default value, as a string or None.
                    Should not be set if the function should raise an exception
                    in cases where no value could be obtained using this URL
                    and name.
    """
    parsed_url = urlparse(url)
    params = dict(parse_qsl(parsed_url.fragment, keep_blank_values=True))
    if name not in params:
        if default is NO_DEFAULT:
            raise ValueError(
                f'URL {url!r} has no fragment parameter named {name!r}.',
            )

        return default

    return params[name]


def get_url_params(url: str) -> Dict[str, str]:
    """Get query parameters from an URL.

    :param url: The URL to get the parameters from.
    """
    parsed_url = urlparse(url)
    return dict(parse_qsl(parsed_url.query, keep_blank_values=True))


def get_url_fragment_params(url: str) -> Dict[str, str]:
    """Get fragment parameters from an URL.

    Note that this function is only for cases where the fragment is encoded
    the same way as a query string, e.g. 'https://example.org/#a=b&c=d'.

    :param url: The URL to get the parameters from.
    """
    parsed_url = urlparse(url)
    return dict(parse_qsl(parsed_url.fragment, keep_blank_values=True))


def get_url_with_params(url: str, **kwargs: str | None) -> str:
    """Get an URL with additional or without some query parameters.

    :param url: The URL to modify.
    """
    parsed_url = urlparse(url)
    params = dict(parse_qsl(parsed_url.query, keep_blank_values=True))

    for key, value in kwargs.items():
        if value is None:
            if key in params:
                del params[key]
        else:
            params[key] = str(value)

    return parsed_url._replace(
        query=urlencode(params, doseq=True),
    ).geturl()



# Copyright(C) 2010-2011 Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict
from configparser import RawConfigParser, DEFAULTSECT
from decimal import Decimal
import os
import io

from .iconfig import IConfig
from .util import LOGGER


__all__ = ['INIConfig']


class INIConfig(IConfig):
    ROOTSECT = 'ROOT'

    def __init__(self, path):
        self.path = path
        self.values = OrderedDict()
        self.config = RawConfigParser()

    def load(self, default={}):
        self.values = OrderedDict(default)

        if os.path.exists(self.path):
            LOGGER.debug('Loading application configuration file: %s.' % self.path)
            self.config.read(self.path, encoding='utf-8')
            for section in self.config.sections():
                args = section.split(':')
                if args[0] == self.ROOTSECT:
                    args.pop(0)
                for key, value in self.config.items(section):
                    self.set(*(args + [key, value]))
            # retro compatibility
            if len(self.config.sections()) == 0:
                first = True
                for key, value in self.config.items(DEFAULTSECT):
                    if first:
                        LOGGER.warning('The configuration file "%s" uses an old-style' % self.path)
                        LOGGER.warning('Please rename the %s section to %s' % (DEFAULTSECT, self.ROOTSECT))
                        first = False
                    self.set(key, value)
            LOGGER.debug('Application configuration file loaded: %s.' % self.path)
        else:
            self.save()
            LOGGER.debug('Application configuration file created with default values: %s. '
                          'Please customize it.' % self.path)
        return self.values

    def save(self):
        def save_section(values, root_section=self.ROOTSECT):
            for k, v in values.items():
                if isinstance(v, (int, Decimal, float, str)):
                    if not self.config.has_section(root_section):
                        self.config.add_section(root_section)
                    self.config.set(root_section, k, str(v))
                elif isinstance(v, dict):
                    new_section = ':'.join((root_section, k)) if (root_section != self.ROOTSECT or k == self.ROOTSECT) else k
                    if not self.config.has_section(new_section):
                        self.config.add_section(new_section)
                    save_section(v, new_section)
        save_section(self.values)
        with io.open(self.path, 'w', encoding='utf-8') as f:
            self.config.write(f)

    def get(self, *args, **kwargs):
        default = None
        if 'default' in kwargs:
            default = kwargs['default']

        v = self.values
        for k in args[:-1]:
            if k in v:
                v = v[k]
            else:
                return default
        try:
            return v[args[-1]]
        except KeyError:
            return default

    def set(self, *args):
        v = self.values
        for k in args[:-2]:
            if k not in v:
                v[k] = OrderedDict()
            v = v[k]
        v[args[-2]] = args[-1]

    def delete(self, *args):
        v = self.values
        for k in args[:-1]:
            if k not in v:
                return
            v = v[k]
        v.pop(args[-1], None)



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict
from datetime import datetime

from dateutil import rrule

from .base import Capability, BaseObject, Field, StringField, BytesField, IntField, \
                  BoolField, UserError
from .address import PostalAddress, compat_field


__all__ = [
    'ProfileNode', 'ContactPhoto', 'Contact', 'QueryError', 'Query', 'CapContact',
    'BaseContact', 'PhysicalEntity', 'Person', 'Place', 'OpeningHours',
    'OpeningRule', 'RRuleField', 'CapDirectory',
]


class ProfileNode:
    """
    Node of a :class:`Contact` profile.
    """
    HEAD =    0x01
    SECTION = 0x02

    def __init__(self, name, label, value, sufix=None, flags=0):
        self.name = name
        self.label = label
        self.value = value
        self.sufix = sufix
        self.flags = flags

    def __getitem__(self, key):
        return self.value[key]


class ContactPhoto(BaseObject):
    """
    Photo of a contact.
    """
    name =              StringField('Name of the photo')
    data =              BytesField('Data of photo')
    thumbnail_url =     StringField('Direct URL to thumbnail')
    thumbnail_data =    BytesField('Data of thumbnail')
    hidden =            BoolField('True if the photo is hidden on website')

    def __init__(self, name, url=None):
        super(ContactPhoto, self).__init__(name, url)
        self.name = name

    def __iscomplete__(self):
        return (self.data and (not self.thumbnail_url or self.thumbnail_data))

    def __str__(self):
        return self.url

    def __repr__(self):
        return '<ContactPhoto %r data=%do tndata=%do>' % (self.id,
                                                          len(self.data) if self.data else 0,
                                                          len(self.thumbnail_data) if self.thumbnail_data else 0)


class BaseContact(BaseObject):
    """
    This is the blase class for a contact.
    """
    name =      StringField('Name of contact')
    phone =     StringField('Phone number')
    email =     StringField('Contact email')
    website =   StringField('Website URL of the contact')


class Advisor(BaseContact):
    """
    An advisor.
    """
    email =     StringField('Mail of advisor')
    phone =     StringField('Phone number of advisor')
    mobile =    StringField('Mobile number of advisor')
    fax =       StringField('Fax number of advisor')
    agency =    StringField('Name of agency')
    address =   StringField('Address of agency')
    role =      StringField('Role of advisor', default="bank")


class Contact(BaseContact):
    """
    A contact.
    """
    STATUS_ONLINE =  0x001
    STATUS_AWAY =    0x002
    STATUS_OFFLINE = 0x004
    STATUS_ALL =     0xfff

    status =        IntField('Status of contact (STATUS_* constants)')
    status_msg =    StringField('Message of status')
    summary =       StringField('Description of contact')
    photos =        Field('List of photos', dict, default=OrderedDict())
    profile =       Field('Contact profile', dict, default=OrderedDict())

    def __init__(self, id, name, status, url=None):
        super(Contact, self).__init__(id, url)
        self.name = name
        self.status = status

    def set_photo(self, name, **kwargs):
        """
        Set photo of contact.

        :param name: name of photo
        :type name: str
        :param kwargs: See :class:`ContactPhoto` to know what other parameters you can use
        """
        if name not in self.photos:
            self.photos[name] = ContactPhoto(name)

        photo = self.photos[name]
        for key, value in kwargs.items():
            setattr(photo, key, value)

    def get_text(self):
        def print_node(node, level=1):
            result = ''
            if node.flags & node.SECTION:
                result += '\t' * level + node.label + '\n'
                for sub in node.value.values():
                    result += print_node(sub, level + 1)
            else:
                if isinstance(node.value, (tuple, list)):
                    value = ', '.join(str(v) for v in node.value)
                elif isinstance(node.value, float):
                    value = '%.2f' % node.value
                else:
                    value = node.value
                result += '\t' * level + '%-20s %s\n' % (node.label + ':', value)
            return result

        result = 'Nickname: %s\n' % self.name
        if self.status & Contact.STATUS_ONLINE:
            s = 'online'
        elif self.status & Contact.STATUS_OFFLINE:
            s = 'offline'
        elif self.status & Contact.STATUS_AWAY:
            s = 'away'
        else:
            s = 'unknown'
        result += 'Status: %s (%s)\n' % (s, self.status_msg)
        result += 'URL: %s\n' % self.url
        result += 'Photos:\n'
        for name, photo in self.photos.items():
            result += '\t%s%s\n' % (photo, ' (hidden)' if photo.hidden else '')
        result += '\nProfile:\n'
        for head in self.profile.values():
            result += print_node(head)
        result += 'Description:\n'
        for s in self.summary.split('\n'):
            result += '\t%s\n' % s
        return result


class QueryError(UserError):
    """
    Raised when unable to send a query to a contact.
    """


class Query(BaseObject):
    """
    Query to send to a contact.
    """
    message =   StringField('Message received')

    def __init__(self, id, message, url=None):
        super(Query, self).__init__(id, url)
        self.message = message


class CapContact(Capability):
    def iter_contacts(self, status=Contact.STATUS_ALL, ids=None):
        """
        Iter contacts

        :param status: get only contacts with the specified status
        :type status: Contact.STATUS_*
        :param ids: if set, get the specified contacts
        :type ids: list[str]
        :rtype: iter[:class:`Contact`]
        """
        raise NotImplementedError()

    def get_contact(self, id):
        """
        Get a contact from his id.

        The default implementation only calls iter_contacts()
        with the proper values, but it might be overloaded
        by backends.

        :param id: the ID requested
        :type id: str
        :rtype: :class:`Contact` or None if not found
        """

        l = self.iter_contacts(ids=[id])
        try:
            return l[0]
        except IndexError:
            return None

    def send_query(self, id):
        """
        Send a query to a contact

        :param id: the ID of contact
        :type id: str
        :rtype: :class:`Query`
        :raises: :class:`QueryError`
        """
        raise NotImplementedError()

    def get_notes(self, id):
        """
        Get personal notes about a contact

        :param id: the ID of the contact
        :type id: str
        :rtype: str
        """
        raise NotImplementedError()

    def save_notes(self, id, notes):
        """
        Set personal notes about a contact

        :param id: the ID of the contact
        :type id: str
        :returns: the str object to save as notes
        """
        raise NotImplementedError()


class PhysicalEntity(BaseContact):
    """
    Contact which has a physical address.
    """
    postal_address = Field('Postal address', PostalAddress)
    address_notes = StringField('Extra address info')

    country = compat_field('postal_address', 'country')
    postcode = compat_field('postal_address', 'postal_code')
    city = compat_field('postal_address', 'city')
    address = compat_field('postal_address', 'street')


class Person(PhysicalEntity):
    pass


class OpeningHours(BaseObject):
    """
    Definition of times when a place is open or closed.

    Consists in a list of :class:`OpeningRule`.
    Rules should be ordered by priority.
    If no rule matches the given date, it is considered closed by default.
    """

    rules = Field('Rules of opening/closing', list)

    def is_open_at(self, query):
        for rule in self.rules:
            if query in rule:
                return rule.is_open

        return False

    @property
    def is_open_now(self):
        return self.is_open_at(datetime.now())


class RRuleField(Field):
    def __init__(self, doc, **kargs):
        super(RRuleField, self).__init__(doc, rrule.rrulebase)

    def convert(self, v):
        if isinstance(v, str):
            return rrule.rrulestr(v)
        return v


class OpeningRule(BaseObject):
    """
    Single rule defining a (recurrent) time interval when a place is open or closed.
    """
    dates = RRuleField('Dates on which this rule applies')
    times = Field('Times of the day this rule applies', list)
    is_open = BoolField('Is it an opening rule or closing rule?')

    def __contains__(self, dt):
        date = dt.date()
        time = dt.time()

        # check times before dates because there are probably fewer entries
        for start, end in self.times:
            if start <= time <= end:
                break
        else:
            return False

        # can't use "date in self.dates" because rrule only matches datetimes
        for occ in self.dates:
            occ = occ.date()
            if occ > date:
                break
            elif occ == date:
                return True

        return False


class Place(PhysicalEntity):
    opening = Field('Opening hours', OpeningHours)


class SearchQuery(BaseObject):
    """
    Parameters to search for contacts.
    """
    name = StringField('Name to search for')
    location = Field('Address where to look', PostalAddress)

    address = compat_field('location', 'street')
    city = compat_field('location', 'city')


class CapDirectory(Capability):
    def search_contacts(self, query, sortby):
        """
        Search contacts matching a query.

        :param query: search parameters
        :type query: :class:`SearchQuery`
        :rtype: iter[:class:`PhysicalEntity`]
        """
        raise NotImplementedError()



# Copyright(C) 2010-2023  Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import os
import sys


if sys.platform == 'win32' \
        or not sys.stdout.isatty() \
        or os.getenv('NO_COLOR') is not None \
        or os.getenv('ANSI_COLORS_DISABLED') is not None:
    #workaround to disable bold
    BOLD   = ''
    NC     = ''         # no color
else:
    BOLD   = '\x1b[1m'
    NC     = '\x1b[0m'


try:
    from termcolor import colored
except ImportError:
    def colored(s, color=None, on_color=None, attrs=None):
        if os.getenv('ANSI_COLORS_DISABLED') is None \
                and os.getenv('NO_COLOR') is None \
                and attrs is not None and 'bold' in attrs:
            return '%s%s%s' % (BOLD, s, NC)
        else:
            return s


__all__ = ['colored', 'NC', 'BOLD']



# Copyright(C) 2012 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import Capability, BaseObject, Field, IntField, DecimalField, \
                  StringField, BytesField, Enum, EnumField, UserError
from .date import DateField
from .address import compat_field, PostalAddress
from .image import BaseImage

__all__ = [
    'CapHousing', 'Housing', 'Query', 'City', 'UTILITIES', 'ENERGY_CLASS',
    'POSTS_TYPES', 'ADVERT_TYPES', 'HOUSE_TYPES', 'TypeNotSupported',
    'HousingPhoto',
]


class TypeNotSupported(UserError):
    """
    Raised when query type is not supported
    """

    def __init__(self,
                 msg='This type of house is not supported by this module'):
        super(TypeNotSupported, self).__init__(msg)


class HousingPhoto(BaseImage):
    """
    Photo of a housing.
    """
    data =      BytesField('Data of photo')

    def __init__(self, url):
        super(HousingPhoto, self).__init__(url.split('/')[-1], url)

    def __str__(self):
        return self.url

    def __repr__(self):
        return '<HousingPhoto %r data=%do url=%r>' % (self.id, len(self.data) if self.data else 0, self.url)


class UTILITIES(Enum):
    INCLUDED = 'C.C.'
    EXCLUDED = 'H.C.'
    UNKNOWN = ''


class ENERGY_CLASS(Enum):
    A = 'A'
    B = 'B'
    C = 'C'
    D = 'D'
    E = 'E'
    F = 'F'
    G = 'G'


class POSTS_TYPES(Enum):
    RENT = 'RENT'
    SALE = 'SALE'
    SHARING = 'SHARING'
    FURNISHED_RENT = 'FURNISHED_RENT'
    VIAGER = 'VIAGER'


class ADVERT_TYPES(Enum):
    PROFESSIONAL = 'Professional'
    PERSONAL = 'Personal'


class HOUSE_TYPES(Enum):
    APART = 'Apartment'
    HOUSE = 'House'
    PARKING = 'Parking'
    LAND = 'Land'
    OTHER = 'Other'
    UNKNOWN = 'Unknown'


class Housing(BaseObject):
    """
    Content of a housing.
    """
    type = EnumField('Type of housing (rent, sale, sharing)',
                     POSTS_TYPES)
    advert_type = EnumField('Type of advert (professional or personal)',
                            ADVERT_TYPES)
    house_type = EnumField('Type of house (apartment, house, parking, …)',
                           HOUSE_TYPES)
    title = StringField('Title of housing')
    area = DecimalField('Area of housing, in m2')
    cost = DecimalField('Cost of housing')
    price_per_meter = DecimalField('Price per meter ratio')
    currency = StringField('Currency of cost')
    utilities = EnumField('Utilities included or not', UTILITIES)
    date = DateField('Date when the housing has been published')
    address = Field('Location of housing', PostalAddress)
    station = StringField('What metro/bus station next to housing')
    text = StringField('Text of the housing')
    phone = StringField('Phone number to contact')
    photos = Field('List of photos', list)
    rooms = DecimalField('Number of rooms')
    bedrooms = DecimalField('Number of bedrooms')
    details = Field('Key/values of details', dict)
    DPE = EnumField('DPE (Energy Performance Certificate)', ENERGY_CLASS)
    GES = EnumField('GES (Greenhouse Gas Emissions)', ENERGY_CLASS)

    location = compat_field('address', 'full_address')


class Query(BaseObject):
    """
    Query to find housings.
    """
    type = EnumField('Type of housing to find (POSTS_TYPES constants)',
                     POSTS_TYPES)
    cities = Field('List of cities to search in', list, tuple)
    area_min = IntField('Minimal area (in m2)')
    area_max = IntField('Maximal area (in m2)')
    cost_min = IntField('Minimal cost')
    cost_max = IntField('Maximal cost')
    nb_rooms = IntField('Number of rooms')
    house_types = Field('List of house types', list, tuple, default=list(HOUSE_TYPES))
    advert_types = Field('List of advert types to filter on', list, tuple,
                         default=list(ADVERT_TYPES))


class City(BaseObject):
    """
    City.
    """
    name = StringField('Name of city')


class CapHousing(Capability):
    """
    Capability of websites to search housings.
    """

    def search_housings(self, query):
        """
        Search housings.

        :param query: search query
        :type query: :class:`Query`
        :rtype: iter[:class:`Housing`]
        """
        raise NotImplementedError()

    def get_housing(self, housing):
        """
        Get an housing from an ID.

        :param housing: ID of the housing
        :type housing: str
        :rtype: :class:`Housing` or None if not found.
        """
        raise NotImplementedError()

    def search_city(self, pattern):
        """
        Search a city from a pattern.

        :param pattern: pattern to search
        :type pattern: str
        :rtype: iter[:class:`City`]
        """
        raise NotImplementedError()



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from codecs import open
import csv

from woob.tools.misc import to_unicode

from .iformatter import IFormatter

__all__ = ['CSVFormatter']


class CSVFormatter(IFormatter):
    def __init__(self, field_separator=";"):
        super(CSVFormatter, self).__init__()
        self.started = False
        self.field_separator = field_separator

    def flush(self):
        self.started = False

    def format_dict(self, item):
        if not isinstance(self.outfile, str):
            return self.write_dict(item, self.outfile)

        with open(self.outfile, "a+", encoding='utf-8') as fp:
            return self.write_dict(item, fp)

    def write_dict(self, item, fp):
        writer = csv.writer(fp, delimiter=self.field_separator)
        if not self.started:
            writer.writerow([to_unicode(v) for v in item.keys()])
            self.started = True

        writer.writerow([str(v) for v in item.values()])



# Copyright(C) 2009-2011  Romain Bignon, Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from email.mime.text import MIMEText
from smtplib import SMTP
from email.header import Header, decode_header
from email.utils import parseaddr, formataddr, formatdate
from email import message_from_file
import time
import re
import logging
import subprocess
import shlex

from woob.core import CallErrors
from woob.capabilities.messages import CapMessages, CapMessagesPost, Thread, Message
from woob.tools.application.repl import ReplApplication
from woob.tools.date import utc2local
from woob.tools.html import html2text
from woob.tools.misc import get_backtrace, to_unicode


__all__ = ['AppSmtp']


class AppSmtp(ReplApplication):
    APPNAME = 'smtp'
    OLD_APPNAME = 'monboob'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Romain Bignon'
    DESCRIPTION = 'Daemon allowing to regularly check for new messages on various websites, ' \
                  'and send an email for each message, and post a reply to a message on a website.'
    SHORT_DESCRIPTION = "daemon to send and check messages"
    CONFIG = {'interval':  300,
              'domain':    'woob.example.org',
              'recipient': 'woob@example.org',
              'smtp':      'localhost',
              'pipe':      '',
              'html':      0}
    CAPS = CapMessages
    DISABLE_REPL = True

    def load_default_backends(self):
        self.load_backends(CapMessages, storage=self.create_storage())

    def main(self, argv):
        self.load_config()
        try:
            self.config.set('interval', int(self.config.get('interval')))
            if self.config.get('interval') < 1:
                raise ValueError()
        except ValueError:
            print('Configuration error: interval must be an integer >0.', file=self.stderr)
            return 1

        try:
            self.config.set('html', int(self.config.get('html')))
            if self.config.get('html') not in (0, 1):
                raise ValueError()
        except ValueError:
            print('Configuration error: html must be 0 or 1.', file=self.stderr)
            return 2

        return super().main(argv)

    def get_email_address_ident(self, msg, header):
        s = msg.get(header)
        if not s:
            return None
        m = re.match('.*<([^@]*)@(.*)>', s)
        if m:
            return m.group(1)
        else:
            try:
                return s.split('@')[0]
            except IndexError:
                return s

    def do_post(self, line):
        """
        post

        Pipe with a mail to post message.
        """
        msg = message_from_file(self.stdin)
        return self.process_incoming_mail(msg)

    def process_incoming_mail(self, msg):
        to = self.get_email_address_ident(msg, 'To')
        sender = msg.get('From')
        reply_to = self.get_email_address_ident(msg, 'In-Reply-To')

        title = msg.get('Subject')
        if title:
            new_title = ''
            for part in decode_header(title):
                if part[1]:
                    new_title += str(part[0], part[1])
                else:
                    new_title += str(part[0])
            title = new_title

        content = ''
        for part in msg.walk():
            if part.get_content_type() == 'text/plain':
                s = part.get_payload(decode=True)
                charsets = part.get_charsets() + msg.get_charsets()
                for charset in charsets:
                    try:
                        if charset is not None:
                            content += str(s, charset)
                        else:
                            content += str(s)
                    except UnicodeError as e:
                        self.logger.warning('Unicode error: %s' % e)
                        continue
                    except Exception as e:
                        self.logger.exception(e)
                        continue
                    else:
                        break

        if len(content) == 0:
            print('Unable to send an empty message', file=self.stderr)
            return 1

        # remove signature
        content = content.split('\n-- \n')[0]

        parent_id = None
        if reply_to is None:
            # This is a new message
            if '.' in to:
                backend_name, thread_id = to.split('.', 1)
            else:
                backend_name = to
                thread_id = None
        else:
            # This is a reply
            try:
                backend_name, id = reply_to.split('.', 1)
                thread_id, parent_id = id.rsplit('.', 1)
            except ValueError:
                print('In-Reply-To header might be in form <backend.thread_id.message_id>', file=self.stderr)
                return 1

            # Default use the To header field to know the backend to use.
            if to and backend_name != to:
                backend_name = to

        try:
            backend = self.woob.backend_instances[backend_name]
        except KeyError:
            print('Backend %s not found' % backend_name, file=self.stderr)
            return 1

        if not backend.has_caps(CapMessagesPost):
            print('The backend %s does not implement CapMessagesPost' % backend_name, file=self.stderr)
            return 1

        thread = Thread(thread_id)
        message = Message(thread,
                          0,
                          title=title,
                          sender=sender,
                          receivers=[to],
                          parent=Message(thread, parent_id) if parent_id else None,
                          content=content)
        try:
            backend.post_message(message)
        except Exception as e:
            content = 'Unable to send message to %s:\n' % thread_id
            content += '\n\t%s\n' % to_unicode(e)
            if logging.root.level <= logging.DEBUG:
                content += '\n%s\n' % to_unicode(get_backtrace(e))
            self.send_email(backend.name, Message(thread,
                                             0,
                                             title='Unable to send message',
                                             sender='AppSmtp',
                                             parent=Message(thread, parent_id) if parent_id else None,
                                             content=content))

    def do_run(self, line):
        """
        run

        Run the fetching daemon.
        """
        self.woob.repeat(self.config.get('interval'), self.process)
        self.woob.loop()

    def do_once(self, line):
        """
        once

        Send mails only once, then exit.
        """
        return self.process()

    def process(self):
        try:
            for message in self.woob.do('iter_unread_messages'):
                if self.send_email(message.backend, message):
                    self.woob[message.backend].set_message_read(message)
        except CallErrors as e:
            self.bcall_errors_handler(e)

    def send_email(self, backend_name, mail):
        domain = self.config.get('domain')
        recipient = self.config.get('recipient')

        parent_message = mail.parent
        references = []
        while parent_message:
            references.append('<%s.%s@%s>' % (backend_name, mail.parent.full_id, domain))
            parent_message = parent_message.parent
        subject = mail.title
        sender = '"%s" <%s@%s>' % (mail.sender.replace('"', '""') if mail.sender else '',
                                    backend_name, domain)

        # assume that .date is an UTC datetime
        date = formatdate(time.mktime(utc2local(mail.date).timetuple()), localtime=True)
        msg_id = '<%s.%s@%s>' % (backend_name, mail.full_id, domain)

        if self.config.get('html') and mail.flags & mail.IS_HTML:
            body = mail.content
            content_type = 'html'
        else:
            if mail.flags & mail.IS_HTML:
                body = html2text(mail.content)
            else:
                body = mail.content
            content_type = 'plain'

        if body is None:
            body = ''

        if mail.signature:
            if self.config.get('html') and mail.flags & mail.IS_HTML:
                body += '<p>-- <br />%s</p>' % mail.signature
            else:
                body += '\n\n-- \n'
                if mail.flags & mail.IS_HTML:
                    body += html2text(mail.signature)
                else:
                    body += mail.signature

        # Header class is smart enough to try US-ASCII, then the charset we
        # provide, then fall back to UTF-8.
        header_charset = 'ISO-8859-1'

        # We must choose the body charset manually
        for body_charset in 'US-ASCII', 'ISO-8859-1', 'UTF-8':
            try:
                body.encode(body_charset)
            except UnicodeError:
                pass
            else:
                break

        # Split real name (which is optional) and email address parts
        sender_name, sender_addr = parseaddr(sender)
        recipient_name, recipient_addr = parseaddr(recipient)

        # We must always pass Unicode strings to Header, otherwise it will
        # use RFC 2047 encoding even on plain ASCII strings.
        sender_name = str(Header(str(sender_name), header_charset))
        recipient_name = str(Header(str(recipient_name), header_charset))

        # Make sure email addresses do not contain non-ASCII characters
        sender_addr = sender_addr.encode('ascii')
        recipient_addr = recipient_addr.encode('ascii')

        # Create the message ('plain' stands for Content-Type: text/plain)
        msg = MIMEText(body.encode(body_charset), content_type, body_charset)
        msg['From'] = formataddr((sender_name, sender_addr))
        msg['To'] = formataddr((recipient_name, recipient_addr))
        msg['Subject'] = Header(str(subject), header_charset)
        msg['Message-Id'] = msg_id
        msg['Date'] = date
        if references:
            msg['In-Reply-To'] = references[0]
            msg['References'] = " ".join(reversed(references))

        self.logger.info('Send mail from <%s> to <%s>' % (sender, recipient))
        if len(self.config.get('pipe')) > 0:
            p = subprocess.Popen(shlex.split(self.config.get('pipe')),
                                 stdin=subprocess.PIPE,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.STDOUT)
            p.stdin.write(msg.as_string())
            p.stdin.close()
            if p.wait() != 0:
                self.logger.error('Unable to deliver mail: %s' % p.stdout.read().strip())
                return False
        else:
            # Send the message via SMTP to localhost:25
            try:
                smtp = SMTP(self.config.get('smtp'))
                smtp.sendmail(sender, recipient, msg.as_string())
            except Exception as e:
                self.logger.error('Unable to deliver mail: %s' % e)
                return False
            else:
                smtp.quit()

        return True



# Copyright(C) 2010-2011 Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.capabilities import UserError


__all__ = ['ResultsCondition', 'ResultsConditionError']


class IResultsCondition:
    def is_valid(self, obj):
        raise NotImplementedError()


class ResultsConditionError(UserError):
    pass


class Condition:
    def __init__(self, left, op, right):
        self.left = left  # Field of the object to test
        self.op = op
        self.right = right


def is_egal(left, right):
    return left == right


def is_notegal(left, right):
    return left != right


def is_sup(left, right):
    return left < right


def is_inf(left, right):
    return left > right


def is_in(left, right):
    return left in right

functions = {'!=': is_notegal, '=': is_egal, '>': is_sup, '<': is_inf, '|': is_in}


class ResultsCondition(IResultsCondition):
    condition_str = None

    # Supported operators
    # =, !=, <, > for float/int/decimal
    # =, != for strings
    # We build a list of list. Return true if each conditions of one list is TRUE
    def __init__(self, condition_str):
        self.limit = None
        or_list = []
        _condition_str = condition_str.split(' LIMIT ')
        if len(_condition_str) == 2:
            try:
                self.limit = int(_condition_str[1])
            except ValueError:
                raise ResultsConditionError('Syntax error in the condition expression, please check documentation')
        condition_str= _condition_str[0]
        for _or in condition_str.split(' OR '):
            and_list = []
            for _and in _or.split(' AND '):
                operator = None
                for op in ['!=', '=', '>', '<', '|']:
                    if op in _and:
                        operator = op
                        break
                if operator is None:
                    raise ResultsConditionError('Could not find a valid operator in sub-expression "%s". Protect the complete condition expression with quotes, or read the documentation in the man manual.' % _and)
                try:
                    l, r = _and.split(operator)
                except ValueError:
                    raise ResultsConditionError('Syntax error in the condition expression, please check documentation')
                and_list.append(Condition(l.strip(), operator, r.strip()))
            or_list.append(and_list)
        self.condition = or_list
        self.condition_str = condition_str

    def is_valid(self, obj):
        import woob.tools.date as date_utils
        import re
        from datetime import date, datetime, timedelta
        d = obj.to_dict()
        # We evaluate all member of a list at each iteration.
        for _or in self.condition:
            myeval = True
            for condition in _or:
                if condition.left in d:
                    # in the case of id, test id@backend and id
                    if condition.left == 'id':
                        tocompare = condition.right
                        evalfullid = functions[condition.op](tocompare, d['id'])
                        evalid = functions[condition.op](tocompare, obj.id)
                        myeval = evalfullid or evalid
                    else:
                        # We have to change the type of v, always gived as string by application
                        typed = type(d[condition.left])
                        try:
                            if isinstance(d[condition.left], date_utils.date):
                                tocompare = date(*[int(x) for x in condition.right.split('-')])
                            elif isinstance(d[condition.left], date_utils.datetime):
                                splitted_datetime = condition.right.split(' ')
                                tocompare = datetime(*([int(x) for x in splitted_datetime[0].split('-')] +
                                                       [int(x) for x in splitted_datetime[1].split(':')]))
                            elif isinstance(d[condition.left], timedelta):
                                time_dict = re.match(r'^\s*((?P<hours>\d+)\s*h)?\s*((?P<minutes>\d+)\s*m)?\s*((?P<seconds>\d+)\s*s)?\s*$',
                                                     condition.right).groupdict()
                                tocompare = timedelta(seconds=int(time_dict['seconds'] or "0"),
                                                      minutes=int(time_dict['minutes'] or "0"),
                                                      hours=int(time_dict['hours'] or "0"))
                            else:
                                tocompare = typed(condition.right)
                            myeval = functions[condition.op](tocompare, d[condition.left])
                        except:
                            myeval = False
                else:
                    raise ResultsConditionError('Field "%s" is not valid.' % condition.left)
                # Do not try all AND conditions if one is false
                if not myeval:
                    break
            # Return True at the first OR valid condition
            if myeval:
                return True
        # If we are here, all OR conditions are False
        return False

    def __str__(self):
        return self.condition_str



# Copyright(C) 2010-2011 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import Capability, BaseObject, BoolField, StringField
from .date import DateField


__all__ = ['Content', 'Revision', 'CapContent']


class Content(BaseObject):
    """
    Content object.
    """
    title =         StringField('Title of content')
    author =        StringField('Original author of content')
    content =       StringField('Body')
    revision =      StringField('ID of revision')


class Revision(BaseObject):
    """
    Revision of a change on a content.
    """
    author =        StringField('Author of revision')
    comment =       StringField('Comment log about revision')
    timestamp =     DateField('Date of revision')
    minor =         BoolField('Is this change minor?')


class CapContent(Capability):
    def get_content(self, id, revision=None):
        """
        Get a content from an ID.

        :param id: ID of content
        :type id: str
        :param revision: if given, get the content at this revision
        :type revision: :class:`Revision`
        :rtype: :class:`Content`
        """
        raise NotImplementedError()

    def iter_revisions(self, id):
        """
        Iter revisions of a content.

        :param id: id of content
        :type id: str
        :rtype: iter[:class:`Revision`]
        """
        raise NotImplementedError()

    def push_content(self, content, message=None, minor=False):
        """
        Push a new revision of a content.

        :param content: object to push
        :type content: :class:`Content`
        :param message: log message to associate to new revision
        :type message: str
        :param minor: this is a minor revision
        :type minor: bool
        """
        raise NotImplementedError()

    def get_content_preview(self, content):
        """
        Get a HTML preview of a content.

        :param content: content object
        :type content: :class:`Content`
        :rtype: str
        """
        raise NotImplementedError()



# Copyright(C) 2010-2011 Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from woob.capabilities.base import NotLoaded, NotAvailable
from .iformatter import IFormatter


__all__ = ['MultilineFormatter']


class MultilineFormatter(IFormatter):
    def __init__(self, key_value_separator=': ', after_item='\n'):
        super(MultilineFormatter, self).__init__()
        self.key_value_separator = key_value_separator
        self.after_item = after_item

    def flush(self):
        pass

    def format_dict(self, item):
        result = '\n'.join('%s%s' % (
            ('%s%s' % (k, self.key_value_separator) if self.display_keys else ''), v)
            for k, v in item.items() if (v is not NotLoaded and v is not NotAvailable))
        if len(item) > 1:
            result += self.after_item
        return result



# Copyright(C) 2010-2020  woob project
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.capabilities.bank.pfm import AccountDiff


__all__ = ['diff_accounts']


def group_by(iterable, func):
    grouped = {}
    for obj in iterable:
        key = func(obj)
        grouped.setdefault(key, []).append(obj)
    return grouped


def match_unique(new_accounts, old_accounts, func):
    # even though it should be unique and maybe required, don't error if it's missing or not unique
    # just ignore bad groups
    new_groups = {key: accs for key, accs in group_by(new_accounts, func).items() if key and len(accs) == 1}
    old_groups = {key: accs for key, accs in group_by(old_accounts, func).items() if key and len(accs) == 1}

    for matching_key in (new_groups.keys() & old_groups.keys()):
        new = new_groups[matching_key][0]
        old = old_groups[matching_key][0]

        new_accounts.remove(new)
        old_accounts.remove(old)
        yield new, old


class IdSet:
    """A set for unhashable objects, expecting to pass the same objects"""
    def __init__(self, elements=()):
        self.container = {}
        for el in elements:
            self.add(el)

    def __iter__(self):
        return iter(self.container.values())

    def __contains__(self, el):
        return id(el) in self.container

    def __len__(self):
        return len(self.container)

    def add(self, el):
        self.container[id(el)] = el

    def remove(self, el):
        del self.container[id(el)]

    def discard(self, el):
        try:
            self.remove(el)
        except KeyError:
            pass


def diff_accounts(backend, new_accounts, old_accounts):
    """Compare accounts between a sync and previous sync

    Tries to match accounts just fetched with accounts fetched from a previous
    run of `iter_accounts()`.

    :param backend: backend from which the objects come
    :type backend: :class:`woob.tools.backend.Module`
    :type new_accounts: iter[:class:`woob.capabilities.bank.Account`]
    :type old_accounts: iter[:class:`woob.capabilities.bank.Account`]
    :rtype: :class:`woob.capabilities.bank.AccountDiff`
    """

    new_accounts = IdSet(new_accounts)
    old_accounts = IdSet(old_accounts)

    diff = AccountDiff()
    diff.matching.extend(match_unique(new_accounts, old_accounts, lambda acc: acc.id))
    diff.matching.extend(match_unique(new_accounts, old_accounts, lambda acc: acc.iban))

    if hasattr(backend, 'diff_accounts'):
        try:
            module_diff = backend.diff_accounts(new_accounts, old_accounts)
        except NotImplementedError:
            pass
        else:
            for new, old in module_diff.matching:
                new_accounts.discard(new)
                old_accounts.discard(old)
                diff.matching.append((new, old))

            diff.obsolete = module_diff.obsolete
            diff.new = module_diff.new
            diff.unknown = module_diff.unknown

    if not new_accounts:
        diff.obsolete = list(old_accounts)
    elif old_accounts:
        diff.unknown = list(new_accounts)
    else:
        diff.new = list(new_accounts)

    return diff



# Copyright(C) 2010-2016 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from decimal import Decimal

from woob.capabilities.base import (
    BaseObject, StringField, DecimalField, Currency,
    Capability,
)
from woob.capabilities.date import DateField


__all__ = [
    'Rate', 'CapCurrencyRate',
]


class Rate(BaseObject, Currency):
    """
    Currency exchange rate.
    """

    currency_from = StringField('The currency to which exchange rates are relative to')
    currency_to = StringField('The currency is converted to')
    value = DecimalField('Exchange rate')
    datetime = DateField('Collection date and time')

    def __repr__(self):
        return "<%s from=%r to=%r value=%r>" % (
            type(self).__name__, self.currency_from, self.currency_to, self.value,
        )

    def convert(self, amount):
        if isinstance(amount, float):
            amount = Decimal(str(amount))
        return amount * self.value


class CapCurrencyRate(Capability):
    """
    Capability of bank websites to get currency exchange rates.
    """

    def iter_currencies(self):
        """
        Iter available currencies.

        :rtype: iter[:class:`Currency`]
        """
        raise NotImplementedError()

    def get_rate(self, currency_from, currency_to):
        """
        Get exchange rate.

        :param currency_from: currency to which exchange rate is relative to
        :type currency_from: :class:`Currency`
        :param currency_to: currency is converted to
        :type currency_to: :class`Currency`
        :rtype: :class:`Rate`
        """
        raise NotImplementedError()



# Copyright(C) 2009-2021  Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import sys
import difflib
import importlib
import pkgutil

import woob
import woob.applications

try:
    import woob_applications
except ImportError:
    woob_applications = None


__all__ = ['Launcher']


class Launcher:
    @classmethod
    def list_apps(cls):
        apps = set()
        for module in pkgutil.iter_modules(woob.applications.__path__):
            apps.add(module.name)

        if woob_applications:
            for module in pkgutil.iter_modules(woob_applications.__path__):
                apps.add(module.name)

        apps.remove("main")
        return sorted(apps)

    @classmethod
    def load_app(cls, app):
        try:
            app_module = importlib.import_module("woob.applications.%s" % app)
        except ImportError:
            if woob_applications is None:
                raise
            app_module = importlib.import_module("woob_applications.%s" % app)

        return getattr(app_module, app_module.__all__[0])

    @classmethod
    def run_app(cls, app, args):
        app_class = cls.load_app(app)
        return app_class.run([app] + args)

    @classmethod
    def print_list(cls, app_list):
        print('usage: woob [--version] <command> [<args>]')
        print()
        print('Use one of this commands:')
        for app in app_list:
            try:
                app_class = cls.load_app(app)
            except ImportError as e:
                print('   %-15s (unable to load: %s)' % (app, e))
            else:
                print('   %-15s %s' % (app_class.APPNAME, app_class.SHORT_DESCRIPTION))
        print()
        print('For more information about a command, use:')
        print('   $ man woob-<command>')
        print('or')
        print('   $ woob <command> --help')

    @classmethod
    def print_version(cls):
        print('%s v%s %s' % (woob.__name__, woob.__version__, woob.__copyright__))

    @classmethod
    def run(cls):
        if sys.version_info < (3, 7):
            print('woob requires python >= 3.7 to work', file=sys.stderr)
            return 1

        app_list = cls.list_apps()

        if len(sys.argv) < 2 or sys.argv[1] == '--help':
            return cls.print_list(app_list)

        if sys.argv[1] == '--version':
            return cls.print_version()

        if sys.argv[1] == 'update':
            sys.argv.insert(1, 'config')

        if sys.argv[1] not in app_list:
            print("woob: '{0}' is not a woob command. See 'woob --help'.".format(sys.argv[1]))
            words = difflib.get_close_matches(sys.argv[1], app_list)
            if words:
                print()
                print('The most similar command is\n\t{0}'.format('\n\t'.join(words)))
            return 1

        return cls.run_app(sys.argv[1], sys.argv[2:])



# Copyright(C) 2010-2023 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import importlib
import posixpath
import shutil
import re
import sys
import os
import subprocess
import hashlib
from compileall import compile_dir
from contextlib import closing, contextmanager
from datetime import datetime
from io import BytesIO, StringIO
from pathlib import Path
from tempfile import NamedTemporaryFile, mkdtemp
from urllib.request import getproxies
from configparser import RawConfigParser, DEFAULTSECT
import tarfile

import packaging.version
from packaging.specifiers import SpecifierSet

from woob.browser.browsers import Browser
from woob.browser.profiles import Woob as WoobProfile
from woob.exceptions import BrowserHTTPError, BrowserHTTPNotFound, ModuleInstallError
from woob.tools.log import getLogger
from woob.tools.misc import get_backtrace, to_unicode, find_exe
from woob.tools.packaging import parse_requirements

from .modules import LoadedModule, _add_in_modules_path


@contextmanager
def open_for_config(filename):
    f = NamedTemporaryFile(
        mode='w', encoding='utf-8', dir=os.path.dirname(filename), delete=False
    )
    with f:
        yield f
    os.replace(f.name, filename)


class ModuleInfo:
    """
    Information about a module available on a repository.
    """

    def __init__(self, name):
        self.name = name

        # path to the local directory containing this module.
        self.path = None
        self.url = None
        self.repo_url = None
        self.signed = False

        self.version = 0
        self.capabilities = ()
        self.dependencies = ()
        self.description = ''
        self.maintainer = ''
        self.license = ''
        self.icon = ''
        self.woob_spec = None

    def load(self, items):
        self.version = int(items['version'])
        self.capabilities = items['capabilities'].split()
        self.dependencies = items.get('dependencies', '').split()
        self.description = to_unicode(items['description'])
        self.maintainer = to_unicode(items['maintainer'])
        self.license = to_unicode(items['license'])
        self.icon = items['icon'].strip() or None
        self.woob_spec = SpecifierSet(items.get('woob_spec', ''))

    def has_caps(self, *caps):
        """Return True if module implements at least one of the caps."""
        if len(caps) == 1 and isinstance(caps[0], (list, tuple)):
            caps = caps[0]
        for c in caps:
            if type(c) == type:
                c = c.__name__
            if c in self.capabilities:
                return True
        return False

    def is_installed(self):
        return self.path is not None

    def is_local(self):
        return self.url is None

    def dump(self):
        return (('version', self.version),
                ('capabilities', ' '.join(self.capabilities)),
                ('dependencies', ' '.join(self.dependencies)),
                ('description', self.description),
                ('maintainer', self.maintainer),
                ('license', self.license),
                ('icon', self.icon or ''),
                ('woob_spec', str(self.woob_spec)),
               )


class RepositoryUnavailable(Exception):
    """
    Repository in not available.
    """


class Repository:
    """
    Represents a repository.
    """
    INDEX = 'modules.list'
    KEYDIR = '.keys'
    KEYRING = 'trusted.gpg'

    def __init__(self, url):
        self.url = url
        self.name = ''
        self.update = 0
        self.maintainer = ''
        self.local = None
        self.signed = False
        self.key_update = 0
        self.obsolete = False
        self.logger = getLogger(f'{__name__}.repository')
        self.errors = {}

        self.modules = {}

        if self.url.startswith('file://'):
            self.local = True
        elif re.match('https?://.*', self.url):
            self.local = False
        else:
            # This is probably a file in ~/.woob/repositories/, we
            # don't know if this is a local or a remote repository.
            with open(self.url, 'r', encoding='utf-8') as fp:
                self.parse_index(fp)

    def __repr__(self):
        return f'<Repository {self.name}>'

    def localurl2path(self):
        """
        Get a local path of a file:// URL.
        """
        assert self.local is True

        if self.url.startswith('file://'):
            return self.url[len('file://'):]
        return self.url

    def retrieve_index(self, browser, repo_path):
        """
        Retrieve the index file of this repository. It can use network
        if this is a remote repository.

        :param repo_path: path to save the downloaded index file (if any).
        :type repo_path: str or None
        """
        built = False
        if self.local:
            # Repository is local, open the file.
            filename = os.path.join(self.localurl2path(), self.INDEX)
            try:
                fp = open(filename, 'r', encoding='utf-8')
            except IOError:
                # This local repository doesn't contain a built modules.list index.
                self.name = Repositories.url2filename(self.url)
                self.build_index(self.localurl2path(), filename)
                built = True
                fp = open(filename, 'r', encoding='utf-8')
        else:
            # This is a remote repository, download file
            try:
                fp = StringIO(browser.open(posixpath.join(self.url, self.INDEX)).text)
            except BrowserHTTPError as e:
                raise RepositoryUnavailable(str(e)) from e

        self.parse_index(fp)
        fp.close()

        # this value can be changed by parse_index
        if self.local and not built:
            # Always rebuild index of a local repository.
            self.build_index(self.localurl2path(), filename)

        # Save the repository index in ~/.woob/repositories/
        if repo_path:
            self.save(repo_path, private=True)

    def retrieve_keyring(self, browser, keyring_path, progress):
        # ignore local
        if self.local:
            return

        keyring = Keyring(keyring_path)
        # prevent previously signed repos from going unsigned
        if not self.signed and keyring.exists():
            raise RepositoryUnavailable('Previously signed repository can not go unsigned')
        if not self.signed:
            return

        if not keyring.exists() or self.key_update > keyring.version:
            # This is a remote repository, download file
            try:
                keyring_data = browser.open(posixpath.join(self.url, self.KEYRING)).content
                sig_data = browser.open(posixpath.join(self.url, self.KEYRING + '.sig')).content
            except BrowserHTTPError as e:
                raise RepositoryUnavailable(str(e)) from e
            if keyring.exists():
                if not keyring.is_valid(keyring_data, sig_data):
                    raise InvalidSignature('the keyring itself')
                progress.progress(0.0, 'The keyring was updated (and validated by the previous one).')
            elif not progress.prompt(
                f'The repository {self.url} isn\'t trusted yet.\n'
                f'Fingerprint of keyring is {hashlib.sha1(keyring_data).hexdigest()}\n'
                'Are you sure you want to continue?'
            ):
                raise RepositoryUnavailable('Repository not trusted')
            keyring.save(keyring_data, self.key_update)
            progress.progress(0.0, str(keyring))

    def parse_index(self, fp):
        """
        Parse index of a repository

        :param fp: file descriptor to read
        :type fp: buffer
        """
        config = RawConfigParser()
        config.read_file(fp)

        # Read default parameters
        items = dict(config.items(DEFAULTSECT))
        try:
            self.name = items['name']
            self.update = int(items['update'])
            self.maintainer = items['maintainer']
            self.signed = bool(int(items.get('signed', '0')))
            self.key_update = int(items.get('key_update', '0'))
            self.obsolete = bool(int(items.get('obsolete', '0')))
        except KeyError as e:
            raise RepositoryUnavailable(f'Missing global parameters in repository: {e}') from e
        except ValueError as e:
            raise RepositoryUnavailable(f'Incorrect value in repository parameters: {e}') from e

        if len(self.name) == 0:
            raise RepositoryUnavailable('Name is empty')

        if 'url' in items:
            self.url = items['url']
            self.local = self.url.startswith('file://')
        elif self.local is None:
            raise RepositoryUnavailable('Missing "url" key in settings')

        # Load modules
        self.modules.clear()
        for section in config.sections():
            module = ModuleInfo(section)
            module.load(dict(config.items(section)))
            if not self.local:
                module.url = posixpath.join(self.url, f'{module.name}.tar.gz')
                module.repo_url = self.url
                module.signed = self.signed
            self.modules[section] = module

    def build_index(self, path, filename):
        """
        Rebuild index of modules of repository.

        :param path: path of the repository
        :type path: str
        :param filename: file to save index
        :type filename: str
        """
        self.logger.debug('Rebuild index')
        self.modules.clear()
        self.errors.clear()

        _add_in_modules_path(path)

        if os.path.isdir(os.path.join(path, self.KEYDIR)):
            self.signed = True
            self.key_update = self.get_tree_mtime(os.path.join(path, self.KEYDIR), True)
        else:
            self.signed = False
            self.key_update = 0

        for name in sorted(os.listdir(path)):
            module_path = os.path.join(path, name)

            # Check for special cases.
            if (
                name.startswith('.')  # ".", "..", and private files and dirs
                or name.startswith('_') or name.endswith('_')
                or name == self.KEYDIR
            ):
                continue

            # Check if the module is indeed a module.
            if os.path.isdir(module_path):
                if not os.path.exists(os.path.join(module_path, '__init__.py')):
                    continue
            else:
                basename = os.path.basename(module_path).casefold()
                if not basename.endswith('.py') or len(basename) < 3:
                    continue

                name = name[:-3]

            try:
                pymodule = importlib.import_module(f'woob_modules.{name}')
                module = LoadedModule(pymodule)
            except Exception as e:  # noqa
                self.logger.warning('Unable to build module %s: [%s] %s', name, type(e).__name__, e)
                bt = get_backtrace(e)
                self.logger.debug(bt)
                self.errors[name] = bt
            else:
                m = ModuleInfo(module.name)
                m.version = self.get_tree_mtime(module_path)
                m.capabilities = list({c.__name__ for c in module.iter_caps()})
                m.dependencies = module.dependencies
                m.description = module.description
                m.maintainer = module.maintainer
                m.license = module.license
                m.icon = module.icon or ''

                module_path = Path(module.path)
                if not os.path.isdir(module_path):
                    module_path = module_path.parent

                requirements = parse_requirements(module_path / 'requirements.txt')
                m.woob_spec = requirements.get('woob', '')

                self.modules[module.name] = m

        self.update = int(datetime.now().strftime('%Y%m%d%H%M'))
        self.save(filename)

    @staticmethod
    def get_tree_mtime(path, include_root=False):
        mtime = 0
        if include_root or not os.path.isdir(path):
            mtime = int(datetime.fromtimestamp(os.path.getmtime(path)).strftime('%Y%m%d%H%M'))
        for root, _, files in os.walk(path):
            for f in files:
                if f.endswith('.pyc'):
                    continue
                m = int(datetime.fromtimestamp(os.path.getmtime(os.path.join(root, f))).strftime('%Y%m%d%H%M'))
                mtime = max(mtime, m)

        return mtime

    def save(self, filename, private=False):
        """
        Save repository into a file (modules.list for example).

        :param filename: path to file to save repository.
        :type filename: str
        :param private: if enabled, save URL of repository.
        :type private: bool
        """
        config = RawConfigParser()
        config.set(DEFAULTSECT, 'name', self.name)
        config.set(DEFAULTSECT, 'update', self.update)
        config.set(DEFAULTSECT, 'maintainer', self.maintainer)
        config.set(DEFAULTSECT, 'signed', int(self.signed))
        config.set(DEFAULTSECT, 'key_update', self.key_update)
        if private:
            config.set(DEFAULTSECT, 'url', self.url)

        for module in self.modules.values():
            config.add_section(module.name)
            for key, value in module.dump():
                config.set(module.name, key, value)

        with open_for_config(filename) as f:
            config.write(f)


class Versions:
    VERSIONS_LIST = 'versions.list'

    def __init__(self, path):
        self.path = path
        self.versions = {}

        config_filename = os.path.join(self.path, self.VERSIONS_LIST)
        try:
            with open(config_filename, 'r', encoding='utf-8') as fp:
                config = RawConfigParser()
                config.read_file(fp, config_filename)

                # Read default parameters
                for key, value in config.items(DEFAULTSECT):
                    self.versions[key] = int(value)
        except IOError:
            pass

    def get(self, name):
        return self.versions.get(name, None)

    def set(self, name, version):
        self.versions[name] = int(version)
        self.save()

    def save(self):
        config = RawConfigParser()
        for name, version in self.versions.items():
            config.set(DEFAULTSECT, name, version)

        with open_for_config(os.path.join(self.path, self.VERSIONS_LIST)) as fp:
            config.write(fp)


class IProgress:
    def progress(self, percent, message):
        raise NotImplementedError()

    def error(self, message):
        raise NotImplementedError()

    def prompt(self, message):
        raise NotImplementedError()

    def __repr__(self):
        return f'<{self.__class__.__name__}>'


class PrintProgress(IProgress):
    def progress(self, percent, message):
        print('=== [%3.0f%%] %s' % (percent*100, message), file=sys.stderr)

    def error(self, message):
        print(f'ERROR: {message}', file=sys.stderr)

    def prompt(self, message):
        print(f'{message} (Y/n): *** ASSUMING YES ***', file=sys.stderr)
        return True


class SubProgress(IProgress):
    def __init__(self, target, steps):
        self.target = target
        self.steps = steps
        self.current = 0

    def progress(self, percent, message):
        return self.target.progress((self.current + percent) / self.steps, message)

    def __getattr__(self, attr):
        return getattr(self.target, attr)


def recursive_deps(direct_deps, key, result=None):
    """
    take a dict of direct dependencies and get all dependencies of an element

    >>> recursive_deps({1: {2, 3}, 2: {3}, 3: {4}, 4: set()}, 1)
    {2, 3, 4}
    """

    if result is None:
        result = set()

    for sub in direct_deps[key]:
        if sub not in result:
            result.add(sub)
            recursive_deps(direct_deps, sub, result)
    return result


class DepList(list):
    def move_value_after(self, val, afters):
        # push an element after all its dependencies

        for pos in range(len(self) - 1, -1, -1):
            if self[pos] == val:
                # already after
                return
            if self[pos] in afters:
                break
        else:
            raise AssertionError()

        current = self.index(val)

        self.insert(pos + 1, val)
        del self[current]


def dependency_sort(deps_rules):
    """
    >>> dependency_sort({1: {2}, 2: {4}, 3: set(), 4: {3}})
    [3, 4, 2, 1]
    """
    # naive but maybe good enough?

    result = DepList(deps_rules)
    deps_rules = dict(deps_rules)

    for key in deps_rules:
        deps = recursive_deps(deps_rules, key)
        result.move_value_after(key, deps)

    return result


DEFAULT_SOURCES_LIST = \
"""# List of woob repositories
#
# The entries below override the entries above (with
# backends of the same name).

https://updates.woob.tech/%(version)s/main/

# DEVELOPMENT
# If you want to hack on woob modules, you may add a
# reference to sources, for example:
#file:///home/rom1/src/woob/woob_modules/
"""


class Repositories:
    SOURCES_LIST = 'sources.list'
    MODULES_DIR = 'modules'
    MODULES_SUBDIR = 'woob_modules'
    REPOS_DIR = 'repositories'
    KEYRINGS_DIR = 'keyrings'
    ICONS_DIR = 'icons'

    SHARE_DIRS = [MODULES_DIR, REPOS_DIR, KEYRINGS_DIR, ICONS_DIR]

    def __init__(self, workdir, datadir, version):
        self.logger = getLogger(f"{__name__}.repositories")
        self.version = version

        self.browser = None

        self.workdir = workdir
        self.datadir = datadir
        self.sources_list = os.path.join(self.workdir, self.SOURCES_LIST)
        self.modules_dir = os.path.join(self.datadir, self.MODULES_DIR, self.version, self.MODULES_SUBDIR)
        self.repos_dir = os.path.join(self.datadir, self.REPOS_DIR)
        self.keyrings_dir = os.path.join(self.datadir, self.KEYRINGS_DIR)
        self.icons_dir = os.path.join(self.datadir, self.ICONS_DIR)

        self.create_dir(self.datadir)
        self.create_dir(self.modules_dir)
        self.create_namespace_package(self.modules_dir)
        self.create_dir(self.repos_dir)
        self.create_dir(self.keyrings_dir)
        self.create_dir(self.icons_dir)

        self.versions = Versions(self.modules_dir)

        self.repositories = []

        if not os.path.exists(self.sources_list):
            with open_for_config(self.sources_list) as f:
                f.write(DEFAULT_SOURCES_LIST)
            self.update()
        else:
            self.load()

    def load_browser(self):
        class WoobBrowser(Browser):
            PROFILE = WoobProfile(self.version)
        if self.browser is None:
            self.browser = WoobBrowser(
                logger=getLogger('browser', parent=self.logger),
                proxy=getproxies())

    def create_dir(self, name):
        if not os.path.exists(name):
            os.makedirs(name)
        elif not os.path.isdir(name):
            self.logger.error('"%s" is not a directory', name)

    namespace_package_content = "from pkgutil import extend_path\n__path__ = extend_path(__path__, __name__)\n"

    def create_namespace_package(self, path):
        pypath = os.path.join(path, '__init__.py')
        if os.path.exists(pypath):
            with open(pypath, 'r', encoding='utf-8') as fd:
                create_file = (fd.read() != self.namespace_package_content)
        else:
            create_file = True

        if create_file:
            with open(pypath, 'wt', encoding='utf-8') as fd:
                fd.write(self.namespace_package_content)

    def _extend_module_info(self, repo, info):
        if repo.local:
            info.path = repo.localurl2path()
        elif self.versions.get(info.name) is not None:
            info.path = self.modules_dir

        return info

    def get_all_modules_info(self, caps=None):
        """
        Get all ModuleInfo instances available.

        :param caps: filter on capabilities:
        :type caps: list[str]
        :rtype: dict[:class:`ModuleInfo`]
        """
        modules = {}
        for repos in reversed(self.repositories):
            for name, info in repos.modules.items():
                if name not in modules and (not caps or info.has_caps(caps)):
                    modules[name] = self._extend_module_info(repos, info)
        return modules

    def get_module_info(self, name):
        """
        Get ModuleInfo object of a module.

        It tries all repositories from last to first, and set
        the 'path' attribute of ModuleInfo if it is installed.
        """
        for repos in reversed(self.repositories):
            if name in repos.modules:
                m = repos.modules[name]
                self._extend_module_info(repos, m)
                return m
        return None

    def load(self):
        """
        Load repositories from ~/.local/share/woob/repositories/.
        """
        self.repositories = []
        for name in sorted(os.listdir(self.repos_dir)):
            path = os.path.join(self.repos_dir, name)
            try:
                repository = Repository(path)
                self.repositories.append(repository)
            except RepositoryUnavailable as e:
                print(f'Unable to load repository {name} ({e}), try to update repositories.', file=sys.stderr)

    def get_module_icon_path(self, module):
        return os.path.join(self.icons_dir, f'{module.name}.png')

    def retrieve_icon(self, module):
        """
        Retrieve the icon of a module and save it in ~/.local/share/woob/icons/.
        """
        self.load_browser()
        if not isinstance(module, ModuleInfo):
            module = self.get_module_info(module)

        dest_path = self.get_module_icon_path(module)

        icon_url = module.icon
        if not icon_url:
            if module.is_local():
                icon_path = os.path.join(module.path, module.name, 'favicon.png')
                if module.path and os.path.exists(icon_path):
                    shutil.copy(icon_path, dest_path)
                return

            icon_url = module.url.replace('.tar.gz', '.png')

        try:
            icon = self.browser.open(icon_url)
        except BrowserHTTPNotFound:
            pass  # no icon, no problem
        else:
            with open(dest_path, 'wb') as fp:
                fp.write(icon.content)

    def _parse_source_list(self):
        l = []
        with open(self.sources_list, 'r', encoding='utf-8') as f:
            for line in f:
                line = line.strip() % {'version': packaging.version.Version(self.version).major}
                m = re.match('(file|https?)://.*', line)
                if m:
                    l.append(line)
        return l

    def update_repositories(self, progress=PrintProgress()):
        """
        Update list of repositories by downloading them
        and put them in ~/.local/share/woob/repositories/.

        :param progress: observer object.
        :type progress: :class:`IProgress`
        """
        self.load_browser()

        self.repositories = []
        for name in os.listdir(self.repos_dir):
            os.remove(os.path.join(self.repos_dir, name))

        gpg_found = Keyring.find_gpg() or Keyring.find_gpgv()
        for line in self._parse_source_list():
            progress.progress(0.0, f'Getting {line}')
            repository = Repository(line)
            filename = self.url2filename(repository.url)
            prio_filename = '%02d-%s' % (len(self.repositories), filename)
            repo_path = os.path.join(self.repos_dir, prio_filename)
            keyring_path = os.path.join(self.keyrings_dir, filename)
            try:
                repository.retrieve_index(self.browser, repo_path)
                if gpg_found:
                    repository.retrieve_keyring(self.browser, keyring_path, progress)
                else:
                    progress.error('Cannot find gpg or gpgv to check for repository authenticity.\n'
                                   'You should install GPG for better security.')
            except RepositoryUnavailable as e:
                progress.error(f'Unable to load repository: {e}')
            else:
                self.repositories.append(repository)
                if repository.obsolete:
                    last_update = datetime.strptime(str(repository.update), '%Y%m%d%H%M').strftime('%Y-%m-%d')
                    progress.error(
                        f'This repository does not receive updates anymore (since {last_update}).\n'
                        'Your woob version is probably obsolete and should be upgraded.'
                    )

    def check_repositories(self):
        """
        Check if sources.list is consistent with repositories
        """
        l = []
        for line in self._parse_source_list():
            repository = Repository(line)
            filename = self.url2filename(repository.url)
            prio_filename = '%02d-%s' % (len(l), filename)
            repo_path = os.path.join(self.repos_dir, prio_filename)
            if not os.path.isfile(repo_path):
                return False
            l.append(repository)
        return True

    def _is_module_updatable(self, info):
        return not info.is_local() and info.is_installed() and self.versions.get(info.name) != info.version

    def _is_module_installable(self, info):
        return not info.is_local() and not info.is_installed()

    def _get_all_dependencies(self, modules):
        modules = list(modules)
        direct_deps = {}

        i = 0
        while i < len(modules):
            current = modules[i]

            for dep_name in current.dependencies:
                if dep_name in direct_deps:
                    # already handled
                    continue

                dep = self.get_module_info(dep_name)
                if not dep:
                    raise ModuleInstallError(f'Module "{dep_name}" does not exist')
                modules.append(dep)

            direct_deps[current.name] = set(current.dependencies)

            i += 1

        # install most depended-on first
        sorted_names = dependency_sort(direct_deps)
        return [self.get_module_info(name) for name in sorted_names]

    def update(self, progress=PrintProgress()):
        """
        Update repositories and install new packages versions.

        :param progress: observer object.
        :type progress: :class:`IProgress`
        """
        self.update_repositories(progress)

        to_update = []
        for info in self.get_all_modules_info().values():
            if self._is_module_updatable(info):
                to_update.append(info)

        to_update = self._get_all_dependencies(to_update)
        to_update = [info for info in to_update if self._is_module_updatable(info) or self._is_module_installable(info)]

        if len(to_update) == 0:
            progress.progress(1.0, 'All modules are up-to-date.')
            return

        proxy_progress = SubProgress(progress, len(to_update))

        for n, info in enumerate(to_update):
            proxy_progress.current = n
            try:
                self._install_one_module(info, proxy_progress)
            except ModuleInstallError as e:
                proxy_progress.progress(1.0, str(e))

    def install(self, module, progress=PrintProgress()):
        if isinstance(module, ModuleInfo):
            info = module
        elif isinstance(module, str):
            progress.progress(0.0, f'Looking for module {module}')
            info = self.get_module_info(module)
            if not info:
                raise ModuleInstallError(f'Module "{module}" does not exist')
        else:
            raise ValueError(f'"module" parameter might be a ModuleInfo object or a string, not {module}')

        to_install = self._get_all_dependencies([info])
        to_install = [
            subinfo
            for subinfo in to_install
            if self._is_module_updatable(subinfo)
            or self._is_module_installable(subinfo)
        ]

        proxy_progress = SubProgress(progress, len(to_install))

        for n, info in enumerate(to_install):
            proxy_progress.current = n
            self._install_one_module(info, proxy_progress)

    def _install_one_module(self, module, progress):
        """
        Install a module.

        :param module: module to install
        :type module: :class:`str` or :class:`ModuleInfo`
        :param progress: observer object
        :type progress: :class:`IProgress`
        """
        if self.version not in module.woob_spec:
            raise ModuleInstallError(
                f"Module requires woob {module.woob_spec}, but you use woob {self.version}'.\n"
                "Hint: use 'woob update' or install a newer version of woob"
            )

        self.load_browser()

        if module.is_local():
            raise ModuleInstallError('%s is available on local.' % module.name)

        module_dir = os.path.join(self.modules_dir, module.name)
        installed = self.versions.get(module.name)
        if installed is None or not os.path.exists(module_dir):
            progress.progress(0.2, 'Module %s is not installed yet' % module.name)
        elif module.version > installed:
            progress.progress(0.2, 'A new version of %s is available' % module.name)
        else:
            raise ModuleInstallError('The latest version of %s is already installed' % module.name)

        progress.progress(0.3, 'Downloading module...')
        try:
            tardata = self.browser.open(module.url).content
        except BrowserHTTPError as e:
            raise ModuleInstallError('Unable to fetch module: %s' % e)

        # Check signature
        if module.signed and (Keyring.find_gpg() or Keyring.find_gpgv()):
            progress.progress(0.5, 'Checking module authenticity...')
            sig_data = self.browser.open(posixpath.join(module.url + '.sig')).content
            keyring_path = os.path.join(self.keyrings_dir, self.url2filename(module.repo_url))
            keyring = Keyring(keyring_path)
            if not keyring.exists():
                raise ModuleInstallError('No keyring found, please update repos.')
            if not keyring.is_valid(tardata, sig_data):
                raise ModuleInstallError('Invalid signature for %s.' % module.name)

        # Extract module from tarball.
        if os.path.isdir(module_dir):
            shutil.rmtree(module_dir)
        progress.progress(0.7, 'Setting up module...')

        # TODO: remove tar.extractall() when not needed to prevent from potential archive attacks
        with closing(tarfile.open('', 'r:gz', BytesIO(tardata))) as tar:
            tar.extractall(self.modules_dir)  # nosec

        if not os.path.isdir(module_dir):
            raise ModuleInstallError(f'The archive for {module.name} looks invalid.')
        # Precompile
        compile_dir(module_dir, quiet=True)

        self.versions.set(module.name, module.version)

        progress.progress(0.9, 'Downloading icon...')
        self.retrieve_icon(module)

        progress.progress(1.0, f'Module {module.name} has been installed!')

    @staticmethod
    def url2filename(url):
        """
        Get a safe file name for an URL.

        All non-alphanumeric characters are replaced by _.
        """
        return ''.join([l if l.isalnum() else '_' for l in url])

    def __iter__(self):
        for repository in self.repositories:
            yield repository

    @property
    def errors(self):
        errors = {}
        for repository in self:
            errors.update(repository.errors)
        return errors


class InvalidSignature(Exception):
    def __init__(self, filename):
        self.filename = filename
        super(InvalidSignature, self).__init__(f'Invalid signature for {filename}')


class Keyring:
    EXTENSION = '.gpg'

    def __init__(self, path):
        self.path = path + self.EXTENSION
        self.vpath = path + '.version'
        self.version = 0

        if self.exists():
            with open(self.vpath, 'r', encoding='utf-8') as f:
                self.version = int(f.read().strip())
        else:
            if os.path.exists(self.path):
                os.remove(self.path)
            if os.path.exists(self.vpath):
                os.remove(self.vpath)

    def exists(self):
        if not os.path.exists(self.vpath):
            return False
        if os.path.exists(self.path):
            # Check the file is not empty.
            # This is because there was a bug creating empty keyring files.
            with open(self.path, 'rb') as fp:
                if len(fp.read().strip()):
                    return True
        return False

    def save(self, keyring_data, version):
        with open(self.path, 'wb') as fp:
            fp.write(keyring_data)
        self.version = version
        with open_for_config(self.vpath) as fp:
            fp.write(str(version))

    @staticmethod
    def find_gpgv():
        return find_exe('gpgv2') or find_exe('gpgv')

    @staticmethod
    def find_gpg():
        return find_exe('gpg2') or find_exe('gpg')

    def is_valid(self, data, sigdata):
        """
        Check if the data is signed by an accepted key.
        data and sigdata should be strings.
        """
        gpg = self.find_gpg()
        gpgv = self.find_gpgv()

        if gpg:
            gpg_homedir = mkdtemp(prefix='woob_gpg_')
            verify_command = [
                gpg, '--verify', '--no-options',
                '--no-default-keyring', '--quiet',
                '--homedir', gpg_homedir
            ]
        elif gpgv:
            verify_command = [gpgv]

        with NamedTemporaryFile(suffix='.sig', delete=False) as sigfile:
            temp_filename = sigfile.name
            return_code = None
            out = ''
            err = ''
            try:
                sigfile.write(sigdata)
                sigfile.flush()  # very important
                sigfile.close()
                assert isinstance(data, bytes)
                # Yes, all of it is necessary
                with subprocess.Popen(verify_command + [
                        '--status-fd', '1',
                        '--keyring', os.path.realpath(self.path),
                        os.path.realpath(sigfile.name),
                        '-'
                    ],
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE) as proc:

                    out, err = proc.communicate(data)
                    return_code = proc.returncode
            finally:
                os.unlink(temp_filename)
                if gpg:
                    shutil.rmtree(gpg_homedir)

            if return_code or b'GOODSIG' not in out or b'VALIDSIG' not in out:
                print(out, err, file=sys.stderr)
                return False
        return True

    def __str__(self):
        if self.exists():
            with open(self.path, 'rb') as f:
                h = hashlib.sha1(f.read()).hexdigest()
            return f'Keyring version {self.version}, checksum {h}'
        return 'NO KEYRING'



# Copyright(C) 2009-2012  Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict
from decimal import Decimal, InvalidOperation
import datetime
import re

from woob.browser.elements import TableElement, ItemElement
from woob.browser.filters.html import TableCell
from woob.browser.filters.standard import Filter, CleanText, CleanDecimal
from woob.capabilities import NotAvailable, NotLoaded
from woob.capabilities.bank import Transaction, Account
from woob.exceptions import ParseError
from woob.tools.date import new_datetime
from woob.tools.log import getLogger
from woob.tools.misc import classproperty


__all__ = [
    'FrenchTransaction', 'AmericanTransaction',
    'sorted_transactions', 'merge_iterators', 'keep_only_card_transactions',
    'omit_deferred_transactions',
]


def parse_with_patterns(raw, obj, patterns):
    obj.label = raw

    for pattern, _type in patterns:
        m = pattern.match(raw)
        if m:
            args = m.groupdict()

            def inargs(key):
                """
                inner function to check if a key is in args,
                and is not None.
                """
                return args.get(key, None) is not None

            obj.type = _type
            labels = [args[name].strip() for name in ('text', 'text2') if inargs(name)]
            if labels:
                obj.label = ' '.join(labels)

            if inargs('category'):
                obj.category = args['category'].strip()

            # Set date from information in raw label.
            if inargs('dd') and inargs('mm'):
                dd = int(args['dd']) if args['dd'] != '00' else 1
                mm = int(args['mm'])

                if inargs('yy'):
                    yy = int(args['yy'])
                else:
                    d = obj.date
                    try:
                        d = d.replace(month=mm, day=dd)
                    except ValueError:
                        d = d.replace(year=d.year-1, month=mm, day=dd)

                    yy = d.year
                    if d > obj.date:
                        yy -= 1

                if yy < 100:
                    yy += 2000

                try:
                    if inargs('HH') and inargs('MM'):
                        obj.rdate = datetime.datetime(yy, mm, dd, int(args['HH']), int(args['MM']))
                    else:
                        obj.rdate = datetime.date(yy, mm, dd)
                except ValueError as e:
                    raise ParseError('Unable to parse date in label %r: %s' % (raw, e))

            break


class FrenchTransaction(Transaction):
    """
    Transaction with some helpers for french bank websites.
    """
    PATTERNS = []

    def __init__(self, id='', *args, **kwargs):
        super(FrenchTransaction, self).__init__(id, *args, **kwargs)
        self._logger = getLogger('%s.FrenchTransaction' % __name__)

    @classmethod
    def clean_amount(klass, text):
        """
        Clean a string containing an amount.
        """
        text = text.replace('.','').replace(',','.')
        return re.sub(r'[^\d\-\.]', '', text)

    def set_amount(self, credit='', debit=''):
        """
        Set an amount value from a string.

        Can take two strings if there are both credit and debit
        columns.
        """
        credit = self.clean_amount(credit)
        debit = self.clean_amount(debit)

        if len(debit) > 0:
            self.amount = - abs(Decimal(debit))
        elif len(credit) > 0:
            self.amount = Decimal(credit)
        else:
            self.amount = Decimal('0')

    def parse_date(self, date):
        if date is None:
            return NotAvailable

        if not isinstance(date, (datetime.date, datetime.datetime)):
            if date.isdigit() and len(date) == 8:
                date = datetime.date(int(date[4:8]), int(date[2:4]), int(date[0:2]))
            elif '/' in date:
                date = datetime.date(*reversed([int(x) for x in date.split('/')]))
        if not isinstance(date, (datetime.date, datetime.datetime)):
            self._logger.warning('Unable to parse date %r' % date)
            date = NotAvailable
        elif date.year < 100:
            date = date.replace(year=2000 + date.year)

        return date

    def parse(self, date, raw, vdate=None):
        r"""
        Parse date and raw strings to create datetime.date objects,
        determine the type of transaction, and create a simplified label

        When calling this method, you should have defined patterns (in the
        PATTERN class attribute) with a list containing tuples of regexp
        and the associated type, for example::

            PATTERNS = [(re.compile(r'^VIR(EMENT)? (?P<text>.*)'), FrenchTransaction.TYPE_TRANSFER),
                        (re.compile(r'^PRLV (?P<text>.*)'),        FrenchTransaction.TYPE_ORDER),
                        (re.compile(r'^(?P<text>.*) CARTE \d+ PAIEMENT CB (?P<dd>\d{2})(?P<mm>\d{2}) ?(.*)$'),
                                                                   FrenchTransaction.TYPE_CARD)
                       ]


        In regexps, you can define this patterns:

            * text: part of label to store in simplified label
            * category: part of label representing the category
            * yy, mm, dd, HH, MM: date and time parts
        """
        self.date = self.parse_date(date)
        self.vdate = self.parse_date(vdate)
        self.rdate = self.date
        self.raw = raw.replace('\n', ' ').strip()

        try:
            parse_with_patterns(self.raw, self, self.PATTERNS)
        except ParseError as e:
            self._logger.warning('Unable to date in label %r: %s' % (self.raw, e))

    @classproperty
    def TransactionElement(k):
        class _TransactionElement(ItemElement):
            klass = k

            obj_date = klass.Date(TableCell('date'))
            obj_vdate = klass.Date(TableCell('vdate', 'date'))
            obj_raw = klass.Raw(TableCell('raw'))
            obj_amount = klass.Amount(TableCell('credit'), TableCell('debit', default=''))

        return _TransactionElement

    @classproperty
    def TransactionsElement(klass):
        class _TransactionsElement(TableElement):
            col_date =       ['Date']
            col_vdate =      ['Valeur']
            col_raw =        ['Opération', 'Libellé', 'Intitulé opération']
            col_credit =     ['Crédit', 'Montant']
            col_debit =      ['Débit']

            item = klass.TransactionElement
        return _TransactionsElement

    class Date(CleanText):
        def __call__(self, item):
            date = super(FrenchTransaction.Date, self).__call__(item)
            return date

        def filter(self, date):
            date = super(FrenchTransaction.Date, self).filter(date)
            if date is None:
                return NotAvailable

            if not isinstance(date, (datetime.date, datetime.datetime)):
                if date.isdigit() and len(date) == 8:
                    date = datetime.date(int(date[4:8]), int(date[2:4]), int(date[0:2]))
                elif '/' in date:
                    date = datetime.date(*reversed([int(x) for x in date.split('/')]))
            if not isinstance(date, (datetime.date, datetime.datetime)):
                date = NotAvailable
            elif date.year < 100:
                date = date.replace(year=2000 + date.year)

            return date

    @classmethod
    def Raw(klass, *args, **kwargs):
        patterns = klass.PATTERNS

        class Filter(CleanText):
            def __call__(self, item):
                raw = super(Filter, self).__call__(item)
                if item.obj.rdate is NotLoaded:
                    item.obj.rdate = item.obj.date

                parse_with_patterns(raw, item.obj, patterns)
                return raw

            def filter(self, text):
                text = super(Filter, self).filter(text)
                return text.replace('\n', ' ').strip()
        return Filter(*args, **kwargs)

    class Currency(CleanText):
        def filter(self, text):
            text = super(FrenchTransaction.Currency, self).filter(text)
            return Account.get_currency(text)

    class Amount(Filter):
        def __init__(self, credit, debit=None, replace_dots=True):
            self.credit_selector = credit
            self.debit_selector = debit
            self.replace_dots = replace_dots

        def __call__(self, item):
            if self.debit_selector:
                try:
                    return - abs(CleanDecimal(self.debit_selector, replace_dots=self.replace_dots)(item))
                except InvalidOperation:
                    pass

            if self.credit_selector:
                try:
                    return CleanDecimal(self.credit_selector, replace_dots=self.replace_dots)(item)
                except InvalidOperation:
                    pass

            return Decimal('0')


class AmericanTransaction(Transaction):
    """
    Transaction with some helpers for american bank websites.
    """
    @classmethod
    def clean_amount(klass, text):
        """
        Clean a string containing an amount.
        """
        # Convert "American" UUU.CC format to "French" UUU,CC format
        if re.search(r'\d\.\d\d(?: [A-Z]+)?$', text):
            text = text.replace(',', ' ').replace('.', ',')
        return FrenchTransaction.clean_amount(text)

    @classmethod
    def decimal_amount(klass, text):
        """
        Convert a string containing an amount to Decimal.
        """
        amnt = AmericanTransaction.clean_amount(text)
        return Decimal(amnt) if amnt else Decimal('0')


def sorted_transactions(iterable):
    """Sort an iterable of transactions in reverse chronological order"""
    return sorted(iterable, reverse=True, key=lambda tr: (tr.date, new_datetime(tr.rdate) if tr.rdate else datetime.datetime.min))


def merge_iterators(*iterables):
    """Merge transactions iterators keeping sort order.

    Each iterator must already be sorted in reverse chronological order.
    """

    def keyfunc(kv):
        return (kv[1].date, kv[1].rdate)

    its = OrderedDict((iter(it), None) for it in iterables)
    for k in list(its):
        try:
            its[k] = next(k)
        except StopIteration:
            del its[k]

    while its:
        k, v = max(its.items(), key=keyfunc)
        yield v

        try:
            its[k] = next(k)
        except StopIteration:
            del its[k]


def keep_only_card_transactions(it, match_func=None):
    """Filter iterator to keep transactions with card types.

    This helper should typically be used when a banking site returns card and non-card
    transactions mixed on the same checking account.

    Types kept are `TYPE_DEFERRED_CARD` and `TYPE_CARD_SUMMARY`.
    Additionally, the amount is inversed for transactions with type `TYPE_CARD_SUMMARY`.
    This is because on the deferred debit card account, summaries should be positive
    as the amount is debitted from checking account to credit the card account.

    The `match_func` can be provided in case of multiple cards, to only return
    transactions of one card.

    :param match_func: optional function to filter transactions further
    :type match_func: callable or None
    """

    for tr in it:
        if tr.type == tr.TYPE_DEFERRED_CARD:
            if match_func is None or match_func(tr):
                yield tr
        elif tr.type == tr.TYPE_CARD_SUMMARY:
            if match_func is None or match_func(tr):
                tr.amount = -tr.amount
                yield tr


def omit_deferred_transactions(it):
    """Filter iterator to omit transactions with type `TYPE_DEFERRED_CARD`.

    This helper should typically be used when a banking site returns card and non-card
    transactions mixed on the same checking account.
    """
    for tr in it:
        if tr.type != tr.TYPE_DEFERRED_CARD:
            yield tr



# Copyright(C) 2013 Bezleputh
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from .calendar import AppCalendar

__all__ = ['AppCalendar']



# Copyright(C) 2016  Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


import re


_country2length = dict(
    AL=28, AD=24, AT=20, AZ=28, BE=16, BH=22, BA=20, BR=29,
    BG=22, CR=21, HR=21, CY=28, CZ=24, DK=18, DO=28, EE=20,
    FO=18, FI=18, FR=27, GE=22, DE=22, GI=23, GR=27, GL=18,
    GT=28, HU=28, IS=26, IE=22, IL=23, IT=27, KZ=20, KW=30,
    LV=21, LB=28, LI=21, LT=20, LU=20, MK=19, MT=31, MR=27,
    MU=30, MC=27, MD=24, ME=22, NL=18, NO=15, PK=24, PS=29,
    PL=28, PT=25, RO=24, SM=27, SA=24, RS=22, SK=24, SI=19,
    ES=24, SE=24, CH=21, TN=24, TR=26, AE=23, GB=22, VG=24,
    MA=28, JO=30, TL=23, XK=20, QA=29,
)

def clean(iban):
    return iban.replace(' ','').replace('\t', '')

def is_iban_valid(iban):
    # Ensure upper alphanumeric input.
    iban = clean(iban)
    if not re.match(r'^[A-Z]{2}\d{2}[\dA-Z]+$', iban):
        return False

    # Validate country code against expected length.
    if iban[:2] in _country2length and len(iban) != _country2length[iban[:2]]:
        return False

    digits = iban2numeric(iban)
    return digits % 97 == 1

def iban2numeric(iban):
    # Shift and convert.
    iban = iban[4:] + iban[:4]
    # BASE 36: 0..9,A..Z -> 0..35
    digits = int(''.join(str(int(ch, 36)) for ch in iban))
    return digits

def find_iban_checksum(iban):
    iban = iban[:2] + '00' + iban[4:]
    digits = str(iban2numeric(iban))
    checksum = 0
    for char in digits:
        checksum *= 10
        checksum += int(char)
        checksum %= 97
    return 98-checksum

def rebuild_iban(iban):
    return iban[:2] + ('%02d' % find_iban_checksum(iban)) + iban[4:]

def rib2iban(rib):
    return rebuild_iban('FR00' + rib)

def find_rib_checksum(bank, counter, account):
    letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    digits = '12345678912345678923456789'
    account = ''.join([(char if char.isdigit() else digits[letters.find(char.upper())]) for char in account])
    rest = (89*int(bank) + 15*int(counter) + 3*int(account)) % 97
    return 97 - rest

def is_rib_valid(rib):
    if len(rib) != 23:
        return False

    return find_rib_checksum(rib[:5], rib[5:10], rib[10:21]) == int(rib[21:23])

def rebuild_rib(rib):
    rib = clean(rib)
    assert len(rib) >= 21
    key = find_rib_checksum(rib[:5], rib[5:10], rib[10:21])
    return rib[:21] + ('%02d' % key)



# Copyright(C) 2010-2023 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import importlib
import logging
import pkgutil
import sys
import warnings
from inspect import getmodule
from pathlib import Path

try:
    from importlib import metadata
except ImportError: # for Python<3.8
    import importlib_metadata as metadata

from packaging.version import Version

from woob import __version__
from woob.exceptions import ModuleLoadError
from woob.tools.backend import Module
from woob.tools.log import getLogger
from woob.tools.packaging import parse_requirements


__all__ = ['LoadedModule', 'ModulesLoader', 'RepositoryModulesLoader']


class LoadedModule:
    def __init__(self, package):
        self.logger = getLogger('woob.backend')
        self.package = package
        self.klass = None

        full_name = package.__name__
        for attrname in dir(self.package):
            attr = getattr(self.package, attrname)

            # Check that the attribute is indeed a 'Module' subclass.
            # Note that we check below if it is indeed defined in the
            # Python module we're importing, so 'Module' itself or
            # 'Module' subclasses imported from other Python modules
            # won't be taken into account.
            try:
                if not issubclass(attr, Module):
                    continue
            except TypeError:
                # Argument 1 must be a class.
                continue

            # Check that the attribute is indeed defined in the loaded
            # Python module specifically.
            module = getmodule(attr)
            if module is None:
                continue

            module_name = module.__name__
            if (
                not module_name.startswith(full_name)
                or module_name[len(full_name):][:1] not in ('', '.')
            ):
                continue

            # Check that there is indeed only one Module subclass defined
            # in the Python module.

            if self.klass is not None:
                raise ImportError(
                    f'At least two modules are defined in "{full_name}": '
                    + f'{attr!r} and {self.klass!r}',
                )

            self.klass = attr

        if not self.klass:
            raise ImportError(
                f'{package} is not a backend (no Module class found)',
            )

    @property
    def name(self):
        return self.klass.NAME

    @property
    def maintainer(self):
        return f'{self.klass.MAINTAINER} <{self.klass.EMAIL}>'

    @property
    def version(self):
        warnings.warn(
            'The LoadedModule.version attribute will be removed.',
            DeprecationWarning,
            stacklevel=2
        )

        return Version(__version__).base_version

    @property
    def description(self):
        return self.klass.DESCRIPTION

    @property
    def license(self):
        return self.klass.LICENSE

    @property
    def config(self):
        return self.klass.CONFIG

    @property
    def website(self):
        if self.klass.BROWSER and hasattr(self.klass.BROWSER, 'BASEURL') and self.klass.BROWSER.BASEURL:
            return self.klass.BROWSER.BASEURL
        if self.klass.BROWSER and hasattr(self.klass.BROWSER, 'DOMAIN') and self.klass.BROWSER.DOMAIN:
            return f'{self.klass.BROWSER.PROTOCOL}://{self.klass.BROWSER.DOMAIN}'

        return None

    @property
    def icon(self):
        return self.klass.ICON

    @property
    def path(self):
        try:
            return self.package.__path__[0]
        except AttributeError:
            # This might yield 'mymodule/__init__.py' instead of 'mymodule'
            # like the previous version, so we keep the first version if avail.
            return getmodule(self.package).__file__

    @property
    def dependencies(self):
        return self.klass.DEPENDENCIES

    def iter_caps(self):
        return self.klass.iter_caps()

    def has_caps(self, *caps):
        """Return True if module implements at least one of the caps."""
        for c in caps:
            if (isinstance(c, str) and c in [cap.__name__ for cap in self.iter_caps()]) or \
               (type(c) == type and issubclass(self.klass, c)):
                return True
        return False

    def create_instance(self, woob, backend_name, config, storage, nofail=False, logger=None):
        backend_instance = self.klass(woob, backend_name, config, storage, logger=logger or self.logger, nofail=nofail)
        self.logger.debug('Created backend "%s" for module "%s"', backend_name, self.name)
        return backend_instance


def _add_in_modules_path(path):
    try:
        import woob_modules
    except ImportError:
        from types import ModuleType

        woob_modules = ModuleType('woob_modules')
        sys.modules['woob_modules'] = woob_modules

        woob_modules.__path__ = [path]
    else:
        if path not in woob_modules.__path__:
            woob_modules.__path__.append(path)


class ModulesLoader:
    """
    Load modules.
    """

    LOADED_MODULE = LoadedModule

    def __init__(self, path=None, version=None):
        self.version = version
        self.path = path
        if self.path:
            _add_in_modules_path(self.path)
        self.loaded = {}
        self.logger = getLogger(f"{__name__}.loader")

    def get_or_load_module(self, module_name):
        """
        Can raise a ModuleLoadError exception.
        """
        if module_name not in self.loaded:
            self.load_module(module_name)
        return self.loaded[module_name]

    def iter_existing_module_names(self):
        try:
            import woob_modules
        except ImportError:
            return

        for module in pkgutil.iter_modules(woob_modules.__path__):
            if module.name.startswith('_') or module.name.endswith('_'):
                continue
            yield module.name

    def module_exists(self, name):
        for existing_module_name in self.iter_existing_module_names():
            if existing_module_name == name:
                return True
        return False

    def load_all(self):
        for existing_module_name in self.iter_existing_module_names():
            try:
                self.load_module(existing_module_name)
            except ModuleLoadError as e:
                self.logger.warning('could not load module %s: %s', existing_module_name, e)

    def load_module(self, module_name):
        module_path = self.get_module_path(module_name)

        if module_name in self.loaded:
            self.logger.debug('Module "%s" is already loaded from %s', module_name, module_path)
            return

        _add_in_modules_path(module_path)

        # Load spec for now to check version without trying to load the module,
        # as if it depends of an uninstalled dependence or a newest version of
        # woob, it may crash.
        module_spec = importlib.util.find_spec(f'woob_modules.{module_name}')
        if module_spec is None:
            raise ModuleLoadError(module_name, f'Module {module_name} does not exist')
        self.check_version(module_name, module_spec)

        try:
            pymodule = importlib.import_module(f'woob_modules.{module_name}')
            module = self.LOADED_MODULE(pymodule)
        except Exception as e:
            if logging.root.level <= logging.DEBUG:
                self.logger.exception(e)
            raise ModuleLoadError(module_name, e) from e

        self.loaded[module_name] = module
        self.logger.debug('Loaded module "%s" from %s' % (
            module.name,
            module.path,
        ))

    def get_module_path(self, module_name):
        return self.path

    def check_version(self, module_name, module_spec):
        woob_version = Version(self.version) if self.version else None

        # For a directory module, module_spec.origin is
        # 'woob_modules/bnp/__init__.py' so get 'woob_modules/bnp'.
        # For a single-file module, module_spec.origin is
        # 'woob_modules/bnp.py' so get 'woob_modules/'.
        # In that case, that's not a problem, we can assume the parent
        # requirements.txt file applies on all single-file modules.
        requirements_path = Path(module_spec.origin).parent / 'requirements.txt'

        for name, spec in parse_requirements(requirements_path).items():
            if name == 'woob':
                if woob_version and woob_version not in spec:
                    # specific user friendly error message
                    raise ModuleLoadError(
                        module_name,
                        f"Module requires woob {spec}, but you use woob {self.version}'.\n"
                        "Hint: use 'woob update' or install a newer version of woob"
                    )
                continue

            try:
                pkg = metadata.distribution(name)
            except metadata.PackageNotFoundError as exc:
                raise ModuleLoadError(
                    module_name,
                    f'Module requires python package "{name}" but not installed.'
                ) from exc

            if Version(pkg.version) not in spec:
                raise ModuleLoadError(
                    module_name,
                    f'Module requires python package "{name}" {spec} but version {pkg.version} is installed'
                )


class RepositoryModulesLoader(ModulesLoader):
    """
    Load modules from repositories.
    """

    def __init__(self, repositories):
        super().__init__(repositories.modules_dir, repositories.version)
        self.repositories = repositories
        # repositories.modules_dir is ...../woob_modules
        # shouldn't be in sys.path, its parent should
        # or we add it in woob_modules.__path__
        # sys.path.append(os.path.dirname(repositories.modules_dir))

    def iter_existing_module_names(self):
        for name in self.repositories.get_all_modules_info():
            yield name

    def get_module_path(self, module_name):
        minfo = self.repositories.get_module_info(module_name)
        if minfo is None:
            raise ModuleLoadError(module_name, f'No such module {module_name}')
        if minfo.path is None:
            raise ModuleLoadError(module_name, f'Module {module_name} is not installed')

        return minfo.path



# Copyright(C) 2013 Bezleputh
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.capabilities.job import CapJob
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import IFormatter, PrettyFormatter

__all__ = ['AppJob']


class JobAdvertFormatter(IFormatter):

    MANDATORY_FIELDS = ('id', 'url', 'publication_date', 'title')

    def format_obj(self, obj, alias):
        result = '%s%s%s\n' % (self.BOLD, obj.title, self.NC)
        result += 'url: %s\n' % obj.url
        if hasattr(obj, 'publication_date') and obj.publication_date:
            result += 'Publication date : %s\n' % obj.publication_date.strftime('%Y-%m-%d')
        if hasattr(obj, 'place') and obj.place:
            result += 'Location: %s\n' % obj.place
        if hasattr(obj, 'society_name') and obj.society_name:
            result += 'Society : %s\n' % obj.society_name
        if hasattr(obj, 'job_name') and obj.job_name:
            result += 'Job name : %s\n' % obj.job_name
        if hasattr(obj, 'contract_type') and obj.contract_type:
            result += 'Contract : %s\n' % obj.contract_type
        if hasattr(obj, 'pay') and obj.pay:
            result += 'Pay : %s\n' % obj.pay
        if hasattr(obj, 'formation') and obj.formation:
            result += 'Formation : %s\n' % obj.formation
        if hasattr(obj, 'experience') and obj.experience:
            result += 'Experience : %s\n' % obj.experience
        if hasattr(obj, 'description') and obj.description:
            result += 'Description : %s\n' % obj.description
        return result


class JobAdvertListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'title')

    def get_title(self, obj):
        return '%s' % (obj.title)

    def get_description(self, obj):
        result = ''
        if hasattr(obj, 'publication_date') and obj.publication_date:
            result += '\tPublication date : %s\n' % obj.publication_date.strftime('%Y-%m-%d')
        if hasattr(obj, 'place') and obj.place:
            result += '\tLocation: %s\n' % obj.place
        if hasattr(obj, 'society_name') and obj.society_name:
            result += '\tSociety : %s\n' % obj.society_name
        if hasattr(obj, 'contract_type') and obj.contract_type:
            result += '\tContract : %s\n' % obj.contract_type
        return result.strip('\n\t')


class AppJob(ReplApplication):
    APPNAME = 'job'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2012-YEAR Bezleputh'
    DESCRIPTION = "Console application to search for a job."
    SHORT_DESCRIPTION = "search for a job"
    CAPS = CapJob
    EXTRA_FORMATTERS = {'job_advert_list': JobAdvertListFormatter,
                        'job_advert': JobAdvertFormatter,
                        }
    COMMANDS_FORMATTERS = {'search': 'job_advert_list',
                           'ls': 'job_advert_list',
                           'info': 'job_advert',
                           }

    @defaultcount(10)
    def do_search(self, pattern):
        """
        search PATTERN

        Search for an advert  matching a PATTERN.
        """
        self.change_path(['search'])
        self.start_format(pattern=pattern)
        for job_advert in self.do('search_job', pattern):
            self.cached_format(job_advert)

    @defaultcount(10)
    def do_ls(self, line):
        """
        advanced search

        Search for an advert matching to advanced filters.
        """
        self.change_path(['advanced'])
        for job_advert in self.do('advanced_search_job'):
            self.cached_format(job_advert)

    def complete_info(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_info(self, _id):
        """
        info ID

        Get information about an advert.
        """
        if not _id:
            print('This command takes an argument: %s' % self.get_command_help('info', short=True), file=self.stderr)
            return 2

        job_advert = self.get_object(_id, 'get_job_advert')

        if not job_advert:
            print('Job advert not found: %s' % _id, file=self.stderr)
            return 3

        self.start_format()
        self.format(job_advert)
